"""2-dimensional representations."""
# pylint: disable=duplicate-code

__all__ = [
    # Base
    "AbstractPos2D",
    "AbstractVel2D",
    "AbstractAcc2D",
    # TwoSphere
    "TwoSpherePos",
    "TwoSphereVel",
    "TwoSphereAcc",
    # Polar
    "PolarPos",
    "PolarVel",
    "PolarAcc",
    # Cartesian
    "CartesianPos2D",
    "CartesianVel2D",
    "CartesianAcc2D",
]

from .base import AbstractAcc2D, AbstractPos2D, AbstractVel2D
from .cartesian import CartesianAcc2D, CartesianPos2D, CartesianVel2D
from .polar import PolarAcc, PolarPos, PolarVel
from .spherical import TwoSphereAcc, TwoSpherePos, TwoSphereVel

# isort: split
from . import (
    compat,  # noqa: F401
    transform,  # noqa: F401
)



"""Representation of coordinates in different systems."""

__all__ = ["IrreversibleDimensionChange"]


class IrreversibleDimensionChange(UserWarning):
    """Raised when a dimension change is irreversible.

    This exception is raised when a dimension change is irreversible.
    For example, changing from Cartesian3D to a Cartesian2D is irreversible
    because the z-component is lost.

    Examples
    --------
    >>> import warnings
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> with warnings.catch_warnings(record=True) as w:
    ...     warnings.simplefilter("always")
    ...     _ = vec.represent_as(cx.CartesianPos2D)
    >>> print(w[0].message)
    irreversible dimension change

    """



"""Transformations between representations."""

__all__: list[str] = []

from typing import Any
from warnings import warn

from plum import dispatch

import quaxed.numpy as jnp
from unxt.quantity import AbstractQuantity, Quantity

from coordinax._src.vectors.d1 import CartesianPos1D, RadialPos
from coordinax._src.vectors.d2 import AbstractPos2D, CartesianPos2D, PolarPos
from coordinax._src.vectors.d3 import (
    AbstractPos3D,
    CartesianPos3D,
    CylindricalPos,
    MathSphericalPos,
    SphericalPos,
)
from coordinax._src.vectors.exceptions import IrreversibleDimensionChange


@dispatch.multi(
    (CartesianPos2D, type[CylindricalPos]),
    (CartesianPos2D, type[SphericalPos]),
    (CartesianPos2D, type[MathSphericalPos]),
)
def represent_as(
    current: AbstractPos2D,
    target: type[AbstractPos3D],
    /,
    z: AbstractQuantity = Quantity(0.0, "m"),
    **kwargs: Any,
) -> AbstractPos3D:
    """AbstractPos2D -> Cartesian2D -> Cartesian3D -> AbstractPos3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos2D.from_([1.0, 2.0], "km")

    >>> x2 = cx.represent_as(x, cx.CylindricalPos, z=u.Quantity(14, "km"))
    >>> x2
    CylindricalPos( rho=Quantity[...](value=f32[], unit=Unit("km")),
                    phi=Angle(value=f32[], unit=Unit("rad")),
                    z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    >>> x3 = cx.represent_as(x, cx.SphericalPos, z=u.Quantity(14, "km"))
    >>> x3
    SphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                  theta=Angle(value=f32[], unit=Unit("rad")),
                  phi=Angle(value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    >>> x3 = cx.represent_as(x, cx.MathSphericalPos, z=u.Quantity(14, "km"))
    >>> x3
    MathSphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                      theta=Angle(value=f32[], unit=Unit("rad")),
                      phi=Angle(value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    """
    cart2 = represent_as(current, CartesianPos2D)
    cart3 = represent_as(cart2, CartesianPos3D, z=z)
    return represent_as(cart3, target)


@dispatch.multi(
    (PolarPos, type[CartesianPos3D]),
)
def represent_as(
    current: AbstractPos2D,
    target: type[AbstractPos3D],
    /,
    z: AbstractQuantity = Quantity(0.0, "m"),
    **kwargs: Any,
) -> AbstractPos3D:
    """AbstractPos2D -> PolarPos -> Cylindrical -> AbstractPos3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.CartesianPos3D, z=u.Quantity(14, "km"))
    >>> x2
    CartesianPos3D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")),
                    z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    polar = represent_as(current, PolarPos)
    cyl = represent_as(polar, CylindricalPos, z=z)
    return represent_as(cyl, target)


# =============================================================================
# CartesianPos2D


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: CartesianPos2D, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """CartesianPos2D -> CartesianPos1D.

    The `y` coordinate is dropped.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos2D.from_([1.0, 2.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D, z=u.Quantity(14, "km"))
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))
    >>> x2.x
    Quantity['length'](Array(1., dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.x)


@dispatch
def represent_as(
    current: CartesianPos2D, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """CartesianPos2D -> RadialPos.

    The `x` and `y` coordinates are converted to the radial coordinate `r`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos2D.from_([1.0, 2.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos, z=u.Quantity(14, "km"))
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))
    >>> x2.r
    Distance(Array(2.236068, dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=jnp.sqrt(current.x**2 + current.y**2))


# -----------------------------------------------
# 3D


@dispatch
def represent_as(
    current: CartesianPos2D,
    target: type[CartesianPos3D],
    /,
    *,
    z: AbstractQuantity = Quantity(0.0, "m"),
    **kwargs: Any,
) -> CartesianPos3D:
    """CartesianPos2D -> CartesianPos3D.

    The `x` and `y` coordinates are converted to the `x` and `y` coordinates of
    the 3D system.  The `z` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos2D.from_([1.0, 2.0], "km")

    >>> x2 = cx.represent_as(x, cx.CartesianPos3D, z=u.Quantity(14, "km"))
    >>> x2
    CartesianPos3D( x=Quantity[...](value=f32[], unit=Unit("km")),
                    y=Quantity[...](value=f32[], unit=Unit("km")),
                    z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    return target(x=current.x, y=current.y, z=z)


# =============================================================================
# PolarPos

# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: PolarPos, target: type[CartesianPos1D], /, **kwargs: Any
) -> CartesianPos1D:
    """PolarPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPos1D)
    >>> x2
    CartesianPos1D(x=Quantity[...](value=f32[], unit=Unit("km")))
    >>> x2.x
    Quantity['length'](Array(0.9848077, dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.r * jnp.cos(current.phi))


@dispatch
def represent_as(
    current: PolarPos, target: type[RadialPos], /, **kwargs: Any
) -> RadialPos:
    """PolarPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPos)
    >>> x2
    RadialPos(r=Distance(value=f32[], unit=Unit("km")))
    >>> x2.r
    Distance(Array(1., dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r)


# -----------------------------------------------
# 3D


@dispatch
def represent_as(
    current: PolarPos,
    target: type[SphericalPos],
    /,
    theta: Quantity["angle"] = Quantity(0.0, "radian"),
    **kwargs: Any,
) -> SphericalPos:
    """PolarPos -> SphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.SphericalPos, theta=u.Quantity(14, "deg"))
    >>> x2
    SphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                  theta=Angle(value=f32[], unit=Unit("deg")),
                  phi=Angle(value=f32[], unit=Unit("deg")) )
    >>> x2.theta
    Angle(Array(14., dtype=float32), unit='deg')

    """
    return target.from_(r=current.r, theta=theta, phi=current.phi)


@dispatch
def represent_as(
    current: PolarPos,
    target: type[MathSphericalPos],
    /,
    phi: Quantity["angle"] = Quantity(0.0, "radian"),
    **kwargs: Any,
) -> MathSphericalPos:
    """PolarPos -> MathSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.MathSphericalPos, phi=u.Quantity(14, "deg"))
    >>> x2
    MathSphericalPos( r=Distance(value=f32[], unit=Unit("km")),
                      theta=Angle(value=f32[], unit=Unit("deg")),
                      phi=Angle(value=f32[], unit=Unit("deg")) )
    >>> x2.phi
    Angle(Array(14., dtype=float32), unit='deg')

    """
    return target.from_(r=current.r, phi=phi, theta=current.phi)


@dispatch
def represent_as(
    current: PolarPos,
    target: type[CylindricalPos],
    /,
    *,
    z: Quantity["length"] = Quantity(0.0, "m"),
    **kwargs: Any,
) -> CylindricalPos:
    """PolarPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.PolarPos(r=u.Quantity(1.0, "km"), phi=u.Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.CylindricalPos, z=u.Quantity(14, "km"))
    >>> x2
    CylindricalPos( rho=Quantity[...](value=f32[], unit=Unit("km")),
                    phi=Angle(value=f32[], unit=Unit("deg")),
                    z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    return target(rho=current.r, phi=current.phi, z=z)



"""Base implementation of coordinate frames."""

__all__ = ["AbstractReferenceFrame"]

from collections.abc import Mapping
from typing import Annotated as Antd, Any
from typing_extensions import Doc

import equinox as eqx
from plum import dispatch

from .api import frame_transform_op
from coordinax._src.operators.base import AbstractOperator


class AbstractReferenceFrame(eqx.Module):  # type: ignore[misc]
    """Base class for all reference frames."""

    # ---------------------------------------------------------------
    # Constructors

    @classmethod
    @dispatch.abstract
    def from_(
        cls: "type[AbstractReferenceFrame]", obj: Any, /
    ) -> "AbstractReferenceFrame":
        """Construct a reference frame."""

    # TODO: examples
    @classmethod
    @dispatch
    def from_(
        cls: "type[AbstractReferenceFrame]", obj: Mapping[str, Any], /
    ) -> "AbstractReferenceFrame":
        """Construct a reference frame from a mapping."""
        return cls(**obj)

    # ---------------------------------------------------------------
    # Transformations

    # TODO: rename?
    def transform_op(
        self, to_frame: Antd["AbstractReferenceFrame", Doc("frame to transform to")], /
    ) -> Antd[AbstractOperator, Doc("frame transform operator")]:
        """Make a frame transform operator.

        Examples
        --------
        >>> import coordinax.frames as cxf

        >>> icrs = cxf.ICRS()
        >>> gcf = cxf.Galactocentric()
        >>> op = icrs.transform_op(gcf)  # ICRS to Galactocentric
        >>> op
        _ICRS2GCFOperator( gcf=Galactocentric( ... ) )

        >>> op = gcf.transform_op(icrs)  # Galactocentric to ICRS
        >>> op
        _GCF2ICRSOperator( gcf=Galactocentric( ... ) )

        """
        return frame_transform_op(self, to_frame)


@AbstractReferenceFrame.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(
    cls: type[AbstractReferenceFrame], obj: AbstractReferenceFrame, /
) -> AbstractReferenceFrame:
    """Construct a reference frame from another reference frame.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> icrs = cxf.ICRS()
    >>> cxf.AbstractReferenceFrame.from_(icrs) is icrs
    True

    >>> try:
    ...     cxf.Galactocentric.from_(icrs)
    ... except TypeError as e:
    ...     print(e)
    Cannot construct 'Galactocentric' from ICRS()

    """
    if not isinstance(obj, cls):
        msg = f"Cannot construct {cls.__qualname__!r} from {obj}"
        raise TypeError(msg)

    return obj



"""Built-in vector classes."""

__all__ = ["SphericalAcc", "SphericalPos", "SphericalVel"]

from functools import partial
from typing import final

import equinox as eqx

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u
from dataclassish.converters import Unless
from unxt.quantity import AbstractQuantity

import coordinax._src.typing as ct
from .base_spherical import (
    AbstractSphericalAcc,
    AbstractSphericalPos,
    AbstractSphericalVel,
    _180d,
    _360d,
)
from coordinax._src.angle import Angle, BatchableAngle
from coordinax._src.distance import AbstractDistance, BatchableDistance, Distance
from coordinax._src.utils import classproperty
from coordinax._src.vectors import checks
from coordinax._src.vectors.converters import converter_azimuth_to_range

##############################################################################
# Position


# TODO: make this an alias for SphericalPolarPos, the more correct description?
@final
class SphericalPos(AbstractSphericalPos):
    """Spherical-Polar coordinates.

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    r : `coordinax.Distance`
        Radial distance r (slant distance to origin),
    theta : `coordinax.angle.Angle`
        Polar angle [0, 180] [deg] where 0 is the z-axis.
    phi : `coordinax.angle.Angle`
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.

    """

    r: BatchableDistance = eqx.field(
        converter=Unless(AbstractDistance, partial(Distance.from_, dtype=float))
    )
    r"""Radial distance :math:`r \in [0,+\infty)`."""

    theta: BatchableAngle = eqx.field(converter=partial(Angle.from_, dtype=float))
    r"""Inclination angle :math:`\theta \in [0,180]`."""

    phi: BatchableAngle = eqx.field(
        converter=Unless(
            Angle, lambda x: converter_azimuth_to_range(Angle.from_(x, dtype=float))
        )
    )
    r"""Azimuthal angle, generally :math:`\phi \in [0,360)`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        checks.check_r_non_negative(self.r)
        checks.check_polar_range(self.theta)

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["SphericalVel"]:
        return SphericalVel


@SphericalPos.from_._f.register  # type: ignore[attr-defined, misc]  # noqa: SLF001
def from_(
    cls: type[SphericalPos],
    *,
    r: AbstractQuantity,
    theta: AbstractQuantity,
    phi: AbstractQuantity,
) -> SphericalPos:
    """Construct SphericalPos, allowing for out-of-range values.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Let's start with a valid input:

    >>> cx.SphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                       theta=u.Quantity(90, "deg"),
    ...                       phi=u.Quantity(0, "deg"))
    SphericalPos(
      r=Distance(value=f32[], unit=Unit("kpc")),
      theta=Angle(value=f32[], unit=Unit("deg")),
      phi=Angle(value=f32[], unit=Unit("deg"))
    )

    The radial distance can be negative, which wraps the azimuthal angle by 180
    degrees and flips the polar angle:

    >>> vec = cx.SphericalPos.from_(r=u.Quantity(-3, "kpc"),
    ...                             theta=u.Quantity(45, "deg"),
    ...                             phi=u.Quantity(0, "deg"))
    >>> vec.r
    Distance(Array(3., dtype=float32), unit='kpc')
    >>> vec.theta
    Angle(Array(135., dtype=float32), unit='deg')
    >>> vec.phi
    Angle(Array(180., dtype=float32), unit='deg')

    The polar angle can be outside the [0, 180] deg range, causing the azimuthal
    angle to be shifted by 180 degrees:

    >>> vec = cx.SphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                             theta=u.Quantity(190, "deg"),
    ...                             phi=u.Quantity(0, "deg"))
    >>> vec.r
    Distance(Array(3., dtype=float32), unit='kpc')
    >>> vec.theta
    Angle(Array(170., dtype=float32), unit='deg')
    >>> vec.phi
    Angle(Array(180., dtype=float32), unit='deg')

    The azimuth can be outside the [0, 360) deg range. This is wrapped to the
    [0, 360) deg range (actually the base from_ does this):

    >>> vec = cx.SphericalPos.from_(r=u.Quantity(3, "kpc"),
    ...                             theta=u.Quantity(90, "deg"),
    ...                             phi=u.Quantity(365, "deg"))
    >>> vec.phi
    Angle(Array(5., dtype=float32), unit='deg')

    """
    # 1) Convert the inputs
    fields = SphericalPos.__dataclass_fields__
    r = fields["r"].metadata["converter"](r)
    theta = fields["theta"].metadata["converter"](theta)
    phi = fields["phi"].metadata["converter"](phi)

    # 2) handle negative distances
    r_pred = r < jnp.zeros_like(r)
    r = qlax.select(r_pred, -r, r)
    phi = qlax.select(r_pred, phi + _180d, phi)
    theta = qlax.select(r_pred, _180d - theta, theta)

    # 3) Handle polar angle outside of [0, 180] degrees
    theta = jnp.mod(theta, _360d)  # wrap to [0, 360) deg
    theta_pred = theta < _180d
    theta = qlax.select(theta_pred, theta, _360d - theta)
    phi = qlax.select(theta_pred, phi, phi + _180d)

    # 4) Construct. This also handles the azimuthal angle wrapping
    return cls(r=r, theta=theta, phi=phi)


##############################################################################


@final
class SphericalVel(AbstractSphericalVel):
    """Spherical differential representation."""

    d_r: ct.BatchableSpeed = eqx.field(
        converter=partial(u.Quantity["speed"].from_, dtype=float)
    )
    r"""Radial speed :math:`dr/dt \in [-\infty, \infty]."""

    d_theta: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(u.Quantity["angular speed"].from_, dtype=float)
    )
    r"""Inclination speed :math:`d\theta/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(u.Quantity["angular speed"].from_, dtype=float)
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[SphericalPos]:
        return SphericalPos

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["SphericalAcc"]:
        return SphericalAcc


##############################################################################


@final
class SphericalAcc(AbstractSphericalAcc):
    """Spherical differential representation."""

    d2_r: ct.BatchableAcc = eqx.field(
        converter=partial(u.Quantity["acceleration"].from_, dtype=float)
    )
    r"""Radial acceleration :math:`d^2r/dt^2 \in [-\infty, \infty]."""

    d2_theta: ct.BatchableAngularAcc = eqx.field(
        converter=partial(u.Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Inclination acceleration :math:`d^2\theta/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=partial(u.Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[SphericalVel]:
        return SphericalVel



"""Space."""

__all__: list[str] = []


from collections.abc import ItemsView, KeysView, Mapping, ValuesView
from dataclasses import Field
from typing import Any

from plum import dispatch

from .core import Space
from coordinax._src.vectors.base import AbstractVector


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `field_items` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def fields(obj: Space, /) -> tuple[Field, ...]:  # type: ignore[type-arg]
    """Return the items from a Space.

    Examples
    --------
    >>> from dataclassish import fields
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)

    >>> fields(space)
    (Field(name='length',type=<class 'coordinax...CartesianPos3D'>,...),
     Field(name='speed',type=<class 'coordinax...CartesianVel3D'>,...),
     Field(name='acceleration',type=<class 'coordinax...CartesianAcc3D'>,...))

    """
    return fields.invoke(Mapping[str, Any])(obj)


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `field_items` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def field_keys(obj: Space, /) -> KeysView[str]:
    """Return the keys from a Space.

    Examples
    --------
    >>> from dataclassish import field_keys
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)

    >>> field_keys(space)
    dict_keys(['length', 'speed', 'acceleration'])

    """
    return obj.keys()


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `field_items` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def field_values(obj: Space, /) -> ValuesView[AbstractVector]:
    """Return the values from a Space.

    Examples
    --------
    >>> from dataclassish import field_values
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)

    >>> field_values(space)
    dict_values([CartesianPos3D(...), CartesianVel3D(...), CartesianAcc3D(...)])

    """
    return obj.values()


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `field_items` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def field_items(obj: Space, /) -> ItemsView[str, AbstractVector]:
    """Return the items from a Space.

    Examples
    --------
    >>> from dataclassish import field_items
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)

    >>> field_items(space)
    dict_items([('length', CartesianPos3D(...)),
                ('speed', CartesianVel3D(...)),
                ('acceleration', CartesianAcc3D(...))])

    """
    return obj.items()


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `replace` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def replace(obj: Space, /, **kwargs: AbstractVector) -> Space:
    """Replace the components of the vector.

    Examples
    --------
    >>> from dataclassish import replace
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> v = cx.CartesianVel3D.from_([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcc3D.from_([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)
    >>> newspace = replace(space, length=cx.CartesianPos3D.from_([3, 2, 1], "km"))
    >>> newspace["length"].x
    Quantity['length'](Array(3., dtype=float32), unit='km')

    """
    return type(obj)(**{**obj, **kwargs})



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanOperator"]


from typing import TYPE_CHECKING, Any, final, overload

import equinox as eqx
from plum import dispatch

import quaxed.numpy as xp
import unxt as u
from dataclassish.converters import Unless

from .base import AbstractGalileanOperator
from .boost import GalileanBoostOperator
from .rotation import GalileanRotationOperator
from .translation import GalileanTranslationOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.composite import AbstractCompositeOperator
from coordinax._src.operators.sequential import OperatorSequence

if TYPE_CHECKING:
    from typing import Self


@final
class GalileanOperator(AbstractCompositeOperator, AbstractGalileanOperator):
    r"""Operator for general Galilean transformations.

    In the transformed frame the coordinates are given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, R \mathbf{x} + \mathbf{v} t + \mathbf{a})

    where :math:`R` is the rotation matrix, :math:`\mathbf{v}` is the boost
    velocity, :math:`\mathbf{a}` is the spatial translationl, and :math:`s` is
    the time translation. This is equivalent to a sequential operation of 1. a
    rotation, 2. a translation, 3. a boost.

    Parameters
    ----------
    translation : `coordinax.operators.GalileanTranslationOperator`
        The spatial translation of the frame. See
        :class:`coordinax.operators.GalileanTranslationOperator` for alternative
        inputs to construct this parameter.
    velocity : :class:`coordinax.operators.GalileanBoostOperator`
        The boost to the frame. See
        :class:`coordinax.operators.GalileanBoostOperator` for alternative
        inputs to construct this parameter.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import coordinax.operators as cxo

    >>> op = cxo.GalileanOperator(
    ...     translation=u.Quantity([0., 2., 3., 4.], "kpc"),
    ...     velocity=u.Quantity([1., 2., 3.], "km/s"))
    >>> op
    GalileanOperator(
      rotation=GalileanRotationOperator(rotation=f32[3,3]),
      translation=GalileanTranslationOperator(
        translation=FourVector(
          t=Quantity[...](value=f32[], unit=Unit("kpc s / km")),
          q=CartesianPos3D( ... ) )
      ),
      velocity=GalileanBoostOperator( velocity=CartesianVel3D( ... ) )
    )

    Note that the translation is a
    :class:`coordinax.operators.GalileanTranslationOperator` with a
    :class:`vector.FourVector` translation, and the velocity is a
    :class:`coordinax.operators.GalileanBoostOperator` with a
    :class:`vector.CartesianVel3D` velocity. We can also construct them
    directly, which allows for other vector types.

    >>> op = cxo.GalileanOperator(
    ...     translation=cxo.GalileanTranslationOperator(
    ...         cx.FourVector(t=u.Quantity(2.5, "Gyr"),
    ...                       q=cx.SphericalPos(r=u.Quantity(1, "kpc"),
    ...                                         theta=u.Quantity(90, "deg"),
    ...                                         phi=u.Quantity(0, "rad") ) ) ),
    ...     velocity=cxo.GalileanBoostOperator(
    ...         cx.CartesianVel3D.from_([1, 2, 3], "km/s") )
    ... )
    >>> op
    GalileanOperator(
      rotation=GalileanRotationOperator(rotation=f32[3,3]),
      translation=GalileanTranslationOperator(
        translation=FourVector(
          t=Quantity[...)](value=f32[], unit=Unit("Gyr")),
          q=SphericalPos( ... )
        )
      ),
      velocity=GalileanBoostOperator( velocity=CartesianVel3D( ... ) )
    )

    Galilean operators can be applied to :class:`vector.FourVector`:

    >>> w = cx.FourVector.from_([0, 0, 0, 0], "kpc")
    >>> new = op(w)
    >>> new
    FourVector(
      t=Quantity[...](value=f32[], unit=Unit("kpc s / km")),
      q=CartesianPos3D( ... )
    )
    >>> new.t.ustrip("Gyr").round(2)
    Array(2.5, dtype=float32)
    >>> new.q.x
    Quantity['length'](Array(3.5567803, dtype=float32), unit='kpc')

    Also the Galilean operators can also be applied to
    :class:`vector.AbstractPos3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "kpc")
    >>> t = u.Quantity(0, "Gyr")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(3.5567803, dtype=float32), unit='kpc')
    >>> newt
    Quantity['time'](Array(2.5, dtype=float32), unit='Gyr')

    """

    rotation: GalileanRotationOperator = eqx.field(
        default=GalileanRotationOperator(xp.eye(3)),
        converter=GalileanRotationOperator.from_,
    )
    """The in-frame spatial rotation."""

    translation: GalileanTranslationOperator = eqx.field(
        default=GalileanTranslationOperator(u.Quantity([0, 0, 0, 0], "kpc")),
        converter=Unless(
            GalileanTranslationOperator, converter=GalileanTranslationOperator.from_
        ),
    )
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters accetps either a
    :class:`coordinax.operators.GalileanTranslationOperator` instance or
    any input that can be used to construct a :meth:`vector.FourVector`, using
    :meth:`vector.FourVector.from_`. See :class:`vector.FourVector` for
    details.
    """

    velocity: GalileanBoostOperator = eqx.field(
        default=GalileanBoostOperator(u.Quantity([0, 0, 0], "km/s")),
        converter=Unless(GalileanBoostOperator, converter=GalileanBoostOperator.from_),
    )
    """The boost to the frame.

    This parameters accepts either a
    :class:`coordinax.operators.GalileanBoostOperator` instance or any input
    that can be used to construct a :class:`vector.CartesianVel3D`, using
    :meth:`vector.CartesianVel3D.from_`. See :class:`vector.CartesianVel3D` for
    details.
    """

    @property
    def operators(
        self,
    ) -> tuple[
        GalileanRotationOperator, GalileanTranslationOperator, GalileanBoostOperator
    ]:
        """Rotation -> translation -> boost."""
        return (self.rotation, self.translation, self.velocity)

    @overload
    def __getitem__(self, key: int) -> AbstractOperator: ...

    @overload
    def __getitem__(self, key: slice) -> "Self": ...

    def __getitem__(self, key: int | slice) -> "AbstractOperator | Self":
        if isinstance(key, int):
            return self.operators[key]
        return OperatorSequence(self.operators[key])


@dispatch  # type: ignore[misc]
def simplify_op(op: GalileanOperator, /, **kwargs: Any) -> AbstractOperator:
    """Simplify a Galilean operator.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax.operators as cxo

    This Galilean operator cannot be simplified:

    >>> op = cxo.GalileanOperator(
    ...     translation=u.Quantity([0., 2., 3., 4.], "kpc"),
    ...     velocity=u.Quantity([1., 2., 3.], "km/s"),
    ...     rotation=jnp.eye(3).at[0, 2].set(1),
    ... )
    >>> op
    GalileanOperator(
      rotation=GalileanRotationOperator(rotation=f32[3,3]),
      translation=GalileanTranslationOperator(
        translation=FourVector(
          t=Quantity[...](value=f32[], unit=Unit("kpc s / km")),
          q=CartesianPos3D( ... ) )
      ),
      velocity=GalileanBoostOperator( velocity=CartesianVel3D( ... ) )
    )

    >>> cxo.simplify_op(op) is op
    True

    This Galilean operator can be simplified in all its components except the
    translation:

    >>> op = cxo.GalileanOperator(translation=u.Quantity([0., 2., 3., 4.], "kpc"))
    >>> cxo.simplify_op(op)
    OperatorSequence(
      operators=( GalileanTranslationOperator( translation=FourVector( ... ) ), )
    )

    """
    simple_ops = [simplify_op(x, **kwargs) for x in op.operators]
    if any(
        not isinstance(x, type(orig))
        for x, orig in zip(simple_ops, op.operators, strict=True)
    ):
        return simplify_op(OperatorSequence(simple_ops))

    return op



"""Built-in vector classes."""

__all__ = [
    "CartesianGeneric3D",
]

from dataclasses import fields
from functools import partial
from typing import TypeVar, final

import equinox as eqx
import jax

import quaxed.numpy as jnp
from unxt.quantity import AbstractQuantity, Quantity

import coordinax._src.typing as ct
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base.mixins import AvalMixin

VT = TypeVar("VT", bound="CartesianGeneric3D")


@final
class CartesianGeneric3D(AvalMixin, AbstractVector):
    """Generic 3D Cartesian coordinates.

    The fields of this class are not restricted to any specific dimensions.

    Examples
    --------
    >>> import coordinax as cx
    >>> vec = cx.CartesianGeneric3D.from_([1, 2, 3], "kg m /s")
    >>> vec
    CartesianGeneric3D(
      x=Quantity[...]( value=f32[], unit=Unit("kg m / s") ),
      y=Quantity[...]( value=f32[], unit=Unit("kg m / s") ),
      z=Quantity[...]( value=f32[], unit=Unit("kg m / s") )
    )

    """

    x: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.from_, dtype=float)
    )

    y: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.from_, dtype=float)
    )

    z: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.from_, dtype=float)
    )

    # -----------------------------------------------------
    # Unary operations

    def __neg__(self) -> "CartesianGeneric3D":
        """Negate the `coordinax.CartesianGeneric3D`.

        Examples
        --------
        >>> import coordinax as cx
        >>> q = cx.CartesianGeneric3D.from_([1, 2, 3], "kpc")
        >>> (-q).x
        Quantity['length'](Array(-1., dtype=float32), unit='kpc')

        """
        return jax.tree.map(jnp.negative, self)


# =====================================================
# Constructors


@CartesianGeneric3D.from_.dispatch  # type: ignore[attr-defined, misc]
def from_(
    cls: type[CartesianGeneric3D],
    obj: AbstractQuantity,  # TODO: Shaped[AbstractQuantity, "*batch 3"]
    /,
) -> CartesianGeneric3D:
    """Construct a 3D Cartesian position.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CartesianGeneric3D.from_(u.Quantity([1, 2, 3], "m"))
    >>> vec
    CartesianGeneric3D(
      x=Quantity[...](value=f32[], unit=Unit("m")),
      y=Quantity[...](value=f32[], unit=Unit("m")),
      z=Quantity[...](value=f32[], unit=Unit("m"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)



"""Built-in vector classes."""

__all__ = ["ProlateSpheroidalAcc", "ProlateSpheroidalPos", "ProlateSpheroidalVel"]

from dataclasses import KW_ONLY
from functools import partial
from typing import final

import equinox as eqx
from jaxtyping import Shaped

import quaxed.numpy as jnp
from dataclassish.converters import Unless
from unxt import Quantity

import coordinax._src.typing as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from coordinax._src.angle import Angle, BatchableAngleQ
from coordinax._src.utils import classproperty
from coordinax._src.vectors.base import VectorAttribute
from coordinax._src.vectors.checks import (
    check_greater_than_equal,
    check_less_than_equal,
    check_non_negative_non_zero,
)
from coordinax._src.vectors.converters import converter_azimuth_to_range


@final
class ProlateSpheroidalPos(AbstractPos3D):
    """Prolate spheroidal coordinates as defined by Dejonghe & de Zeeuw 1988.

    Note that valid coordinates have:
    - mu >= Delta^2
    - |nu| <= Delta^2
    - Delta > 0

    Parameters
    ----------
    mu : Quantity["area"]
        The spheroidal mu coordinate. This is called `lambda` by Dejonghe & de Zeeuw.
    nu : Quantity["area"]
        The spheroidal nu coordinate.
    phi : `coordinax.angle.Angle`
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.
    Delta : Quantity["length"]
        The focal length of the coordinate system. Must be > 0.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(3.0, "kpc2"),
    ...     nu=Quantity(0.5, "kpc2"),
    ...     phi=Quantity(0.25, "rad"),
    ...     Delta=Quantity(1.5, "kpc"),
    ... )
    >>> vec
    ProlateSpheroidalPos(
      mu=Quantity[PhysicalType('area')](value=f32[], unit=Unit("kpc2")),
      nu=Quantity[PhysicalType('area')](value=f32[], unit=Unit("kpc2")),
      phi=Angle(value=f32[], unit=Unit("rad")),
      Delta=Quantity[PhysicalType('length')](value=weak_f32[], unit=Unit("kpc"))
    )

    This fails with a zero or negative Delta:

    >>> try: vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(3.0, "kpc2"),
    ...     nu=Quantity(0.5, "kpc2"),
    ...     phi=Quantity(0.25, "rad"),
    ...     Delta=Quantity(0.0, "kpc"),
    ... )
    ... except Exception as e: pass

    Or with invalid mu and nu:

    >>> try: vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(0.5, "kpc2"),
    ...     nu=Quantity(0.5, "kpc2"),
    ...     phi=Quantity(0.25, "rad"),
    ...     Delta=Quantity(1.5, "kpc"),
    ... )
    ... except Exception as e: pass

    """

    mu: ct.BatchableArea = eqx.field(
        converter=partial(Quantity["area"].from_, dtype=float)
    )
    r"""Spheroidal mu coordinate :math:`\mu \in [0,+\infty)` (called :math:`\lambda` in
     some Galactic dynamics contexts)."""

    nu: ct.BatchableArea = eqx.field(
        converter=partial(Quantity["area"].from_, dtype=float)
    )
    r"""Spheroidal nu coordinate :math:`\lambda \in [-\infty,+\infty)`."""

    phi: BatchableAngleQ = eqx.field(
        converter=Unless(
            Angle, lambda x: converter_azimuth_to_range(Angle.from_(x, dtype=float))
        )
    )
    r"""Azimuthal angle, generally :math:`\phi \in [0,360)`."""

    _: KW_ONLY
    Delta: Shaped[Quantity["length"], ""] = VectorAttribute()
    """Focal length of the coordinate system."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        check_non_negative_non_zero(self.Delta, name="Delta")
        check_greater_than_equal(
            self.mu, self.Delta**2, name="mu", comparison_name="Delta^2"
        )
        check_less_than_equal(
            jnp.abs(self.nu), self.Delta**2, name="nu", comparison_name="Delta^2"
        )

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["ProlateSpheroidalVel"]:
        return ProlateSpheroidalVel


@final
class ProlateSpheroidalVel(AbstractVel3D):
    """Prolate spheroidal differential representation."""

    d_mu: ct.BatchableDiffusivity = eqx.field(
        converter=partial(Quantity["diffusivity"].from_, dtype=float)
    )
    r"""Prolate spheroidal mu speed :math:`d\mu/dt \in [-\infty, \infty]."""

    d_nu: ct.BatchableDiffusivity = eqx.field(
        converter=partial(Quantity["diffusivity"].from_, dtype=float)
    )
    r"""Prolate spheroidal nu speed :math:`d\nu/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].from_, dtype=float)
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[ProlateSpheroidalPos]:
        return ProlateSpheroidalPos

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["ProlateSpheroidalAcc"]:
        return ProlateSpheroidalAcc


@final
class ProlateSpheroidalAcc(AbstractAcc3D):
    """Prolate spheroidal acceleration representation."""

    d2_mu: ct.BatchableSpecificEnergy = eqx.field(
        converter=partial(Quantity["specific energy"].from_, dtype=float)
    )
    r"""Prolate spheroidal mu acceleration :math:`d^2\mu/dt^2 \in [-\infty, \infty]."""

    d2_nu: ct.BatchableSpecificEnergy = eqx.field(
        converter=partial(Quantity["specific energy"].from_, dtype=float)
    )
    r"""Prolate spheroidal nu acceleration :math:`d^2\nu/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=partial(Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[ProlateSpheroidalVel]:
        return ProlateSpheroidalVel



"""Base classes for operators on coordinates and potentials."""

__all__ = ["simplify_op"]

from typing import Any

from plum import dispatch


@dispatch.abstract  # type: ignore[misc]
def simplify_op(op: Any, /) -> Any:
    """Simplify an operator."""



"""Test using Jax operations."""

import equinox as eqx
import pytest

from dataclassish import field_items
from unxt.quantity import AbstractQuantity, Quantity

import coordinax as cx
from coordinax._src.vectors.base.base_pos import POSITION_CLASSES

POSITION_CLASSES_3D = [c for c in POSITION_CLASSES if issubclass(c, cx.AbstractPos3D)]


# TODO: cycle through all representations
@pytest.fixture(params=POSITION_CLASSES_3D)
def q(request) -> cx.AbstractPos:
    """Fixture for 3D Vectors."""
    q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")

    # Special case ProlateSpheroidalPos, which requires a value of Delta to define the
    # coordinate system
    kwargs = (
        {}
        if request.param is not cx.ProlateSpheroidalPos
        else {"Delta": Quantity(1.0, "kpc")}
    )

    return q.represent_as(request.param, **kwargs)


@eqx.filter_jit
def func(q: cx.AbstractPos, target: type[cx.AbstractPos]) -> cx.AbstractPos:
    # Special case ProlateSpheroidalPos, which requires a value of Delta to define the
    # coordinate system
    kwargs = (
        {} if target is not cx.ProlateSpheroidalPos else {"Delta": Quantity(1.0, "kpc")}
    )

    return q.represent_as(target, **kwargs)


@pytest.mark.parametrize("target", POSITION_CLASSES_3D)
def test_jax_through_representation(
    q: cx.AbstractPos, target: type[cx.AbstractPos]
) -> None:
    """Test using Jax operations through representation."""
    newq = func(q, target)

    assert isinstance(newq, cx.AbstractPos)
    for k, f in field_items(newq):
        assert isinstance(f, AbstractQuantity), k



"""Distance quantities."""

__all__ = ["Distance", "DistanceModulus", "Parallax"]

from dataclasses import KW_ONLY
from typing import Any, final

import equinox as eqx
import jax.numpy as jnp

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractDistance

parallax_base_length = u.Quantity(1, "AU")
distance_modulus_base_distance = u.Quantity(10, "pc")
angle_dimension = u.dimension("angle")
length_dimension = u.dimension("length")


##############################################################################


@final
class Distance(AbstractDistance):
    """Distance quantities.

    The distance is a quantity with dimensions of length.

    Examples
    --------
    >>> from coordinax.distance import Distance
    >>> Distance(10, "pc")
    Distance(Array(10, dtype=int32, ...), unit='pc')

    The units are checked to have length dimensions.

    >>> try: Distance(10, "s")
    ... except ValueError as e: print(e)
    Distance must have dimensions length.

    """

    def __check_init__(self) -> None:
        """Check the initialization."""
        if u.dimension_of(self) != length_dimension:
            msg = "Distance must have dimensions length."
            raise ValueError(msg)

    @property
    def distance(self) -> "Distance":
        """The distance.

        Examples
        --------
        >>> from coordinax.distance import Distance
        >>> d = Distance(10, "pc")
        >>> d.distance is d
        True

        """
        return self

    @property
    def parallax(  # noqa: PLR0206  (needed for quax boundary)
        self, base_length: u.Quantity["length"] = parallax_base_length
    ) -> "Parallax":
        r"""The parallax of the distance.

        The parallax is calculated as :math:`\arctan(1 AU / d)`.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> from coordinax.distance import Distance
        >>> d = Distance(1, "pc")
        >>> jnp.round(d.parallax.to("arcsec"), 2)
        Parallax(Array(1., dtype=float32, ...), unit='arcsec')

        """
        v = jnp.arctan2(base_length, self)
        return Parallax(v.value, v.unit)

    @property
    def distance_modulus(  # noqa: PLR0206  (needed for quax boundary)
        self, base_length: u.Quantity["length"] = distance_modulus_base_distance
    ) -> "DistanceModulus":
        """The distance modulus.

        Examples
        --------
        >>> from coordinax.distance import Distance
        >>> d = Distance(10, "pc")
        >>> d.distance_modulus
        DistanceModulus(Array(0., dtype=float32), unit='mag')

        """
        return DistanceModulus(5 * jnp.log10(self / base_length).value, "mag")


##############################################################################


@final
class Parallax(AbstractDistance):
    """Parallax distance quantity.

    Examples
    --------
    >>> from coordinax.distance import Parallax
    >>> Parallax(1, "mas")
    Parallax(Array(1, dtype=int32, ...), unit='mas')

    The units are checked to have angle dimensions.

    >>> try: Parallax(1, "pc")
    ... except ValueError as e: print(e)
    Parallax must have angular dimensions.

    The parallax is checked to be non-negative by default.

    >>> try: Parallax(-1, "mas")
    ... except Exception: print("negative")
    negative

    To disable this check, set `check_negative=False`.

    >>> Parallax(-1, "mas", check_negative=False)
    Parallax(Array(-1, dtype=int32, ...), unit='mas')

    """

    _: KW_ONLY
    check_negative: bool = eqx.field(default=True, static=True, compare=False)
    """Whether to check that the parallax is strictly non-negative.

    Theoretically the parallax must be strictly non-negative (:math:`\tan(p) = 1
    AU / d`), however noisy direct measurements of the parallax can be negative.
    """

    def __check_init__(self) -> None:
        """Check the initialization."""
        if u.dimension_of(self) != angle_dimension:
            msg = "Parallax must have angular dimensions."
            raise ValueError(msg)

        if self.check_negative:
            eqx.error_if(
                self.value,
                jnp.any(jnp.less(self.value, 0)),
                "Parallax must be non-negative.",
            )

    @property
    def distance(  # noqa: PLR0206  (needed for quax boundary)
        self, base_length: u.Quantity["length"] = parallax_base_length
    ) -> Distance:
        r"""The distance.

        The distance is calculated as :math:`1 AU / \tan(p)`.

        Examples
        --------
        >>> from coordinax.distance import Parallax
        >>> p = Parallax(1, "mas")
        >>> p.distance.to("kpc")
        Distance(Array(1., dtype=float32, ...), unit='kpc')

        """
        v = base_length / jnp.tan(self)
        return Distance(v.value, v.unit)

    @property
    def parallax(self) -> "Parallax":
        """The parallax of the distance.

        Examples
        --------
        >>> from coordinax.distance import Parallax
        >>> p = Parallax(1, "mas")
        >>> p.parallax is p
        True

        """
        return self

    @property
    def distance_modulus(self) -> "DistanceModulus":
        """The distance modulus.

        Examples
        --------
        >>> from coordinax.distance import Parallax
        >>> Parallax(1, "mas").distance_modulus
        DistanceModulus(Array(10., dtype=float32), unit='mag')

        """
        return self.distance.distance_modulus  # TODO: specific shortcut


##############################################################################


@final
class DistanceModulus(AbstractDistance):
    """Distance modulus quantity.

    Examples
    --------
    >>> from coordinax.distance import DistanceModulus
    >>> DistanceModulus(10, "mag")
    DistanceModulus(Array(10, dtype=int32, ...), unit='mag')

    The units are checked to have magnitude dimensions.

    >>> try: DistanceModulus(10, "pc")
    ... except ValueError as e: print(e)
    Distance modulus must have units of magnitude.

    """

    def __check_init__(self) -> None:
        """Check the initialization."""
        if self.unit != u.unit("mag"):
            msg = "Distance modulus must have units of magnitude."
            raise ValueError(msg)

    @property
    def distance(self) -> Distance:
        """The distance.

        The distance is calculated as :math:`10^{(m / 5 + 1)}`.

        Examples
        --------
        >>> from coordinax.distance import DistanceModulus
        >>> DistanceModulus(10, "mag").distance
        Distance(Array(1000., dtype=float32, ...), unit='pc')

        """
        return Distance(10 ** (self.value / 5 + 1), "pc")

    @property
    def parallax(self) -> Parallax:
        """The parallax.

        Examples
        --------
        >>> from coordinax.distance import DistanceModulus
        >>> DistanceModulus(10, "mag").parallax.to("mas")
        Parallax(Array(0.99999994, dtype=float32, ...), unit='mas')

        """
        return self.distance.parallax  # TODO: specific shortcut

    @property
    def distance_modulus(self) -> "DistanceModulus":
        """The distance modulus.

        Examples
        --------
        >>> from coordinax.distance import DistanceModulus
        >>> DistanceModulus(10, "mag").distance_modulus
        DistanceModulus(Array(10, dtype=int32, ...), unit='mag')

        """
        return self


##############################################################################
# Additional constructors


@Distance.from_._f.register  # noqa: SLF001
def from_(
    cls: type[Distance], value: Parallax | u.Quantity["angle"], /, *, dtype: Any = None
) -> Distance:
    """Construct a `coordinax.Distance` from an angle through the parallax.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Distance, Parallax

    >>> Distance.from_(Parallax(1, "mas")).to("kpc")
    Distance(Array(1., dtype=float32, ...), unit='kpc')

    >>> Distance.from_(u.Quantity(1, "mas")).to("kpc")
    Distance(Array(1., dtype=float32, ...), unit='kpc')

    """
    d = parallax_base_length / jnp.tan(value)
    return cls(jnp.asarray(d.value, dtype=dtype), d.unit)


@Distance.from_._f.register  # type: ignore[no-redef]  # noqa: SLF001
def from_(
    cls: type[Distance],
    value: DistanceModulus | u.Quantity["mag"],
    /,
    *,
    dtype: Any = None,
) -> Distance:
    """Construct a `Distance` from a mag through the dist mod.

    Examples
    --------
    >>> import unxt as u
    >>> from coordinax.distance import Distance, DistanceModulus

    >>> Distance.from_(DistanceModulus(10, "mag")).to("pc")
    Distance(Array(1000., dtype=float32, ...), unit='pc')

    >>> Distance.from_(u.Quantity(10, "mag")).to("pc")
    Distance(Array(1000., dtype=float32, ...), unit='pc')

    """
    d = 10 ** (u.ustrip("mag", value) / 5 + 1)
    return cls(jnp.asarray(d, dtype=dtype), "pc")


@Parallax.from_._f.register  # type: ignore[no-redef]  # noqa: SLF001
def from_(
    cls: type[Parallax], value: Distance | u.Quantity["length"], /, *, dtype: Any = None
) -> Parallax:
    """Construct a `Parallax` from a distance.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> from coordinax.distance import Parallax, Distance

    >>> jnp.round(Parallax.from_(Distance(1, "pc")).to("mas"))
    Parallax(Array(1000., dtype=float32, ...), unit='mas')

    >>> jnp.round(Parallax.from_(u.Quantity(1, "pc")).to("mas"), 2)
    Parallax(Array(1000., dtype=float32, ...), unit='mas')

    """
    p = jnp.atan2(parallax_base_length, value)
    return cls(jnp.asarray(p.value, dtype=dtype), p.unit)



"""Representation of coordinates in different systems."""
# ruff: noqa: N803, N806

__all__: list[str] = []

from typing import Any

import equinox as eqx
import jax
from plum import dispatch

import quaxed.numpy as xp
import unxt as u

from .base import AbstractPos3D, AbstractVel3D
from .base_spherical import AbstractSphericalPos
from .cartesian import CartesianAcc3D, CartesianPos3D, CartesianVel3D
from .cylindrical import CylindricalPos, CylindricalVel
from .lonlatspherical import (
    LonCosLatSphericalVel,
    LonLatSphericalPos,
    LonLatSphericalVel,
)
from .mathspherical import MathSphericalPos, MathSphericalVel
from .spherical import SphericalPos, SphericalVel
from .spheroidal import ProlateSpheroidalPos, ProlateSpheroidalVel
from coordinax._src.vectors.base import AbstractPos

###############################################################################
# 3D


@dispatch
def represent_as(
    current: AbstractPos3D, target: type[AbstractPos3D], /, **kwargs: Any
) -> AbstractPos3D:
    """AbstractPos3D -> Cartesian3D -> AbstractPos3D."""
    return represent_as(represent_as(current, CartesianPos3D), target)


@dispatch.multi(
    (CartesianPos3D, type[CartesianPos3D]),
    (CylindricalPos, type[CylindricalPos]),
    (SphericalPos, type[SphericalPos]),
    (LonLatSphericalPos, type[LonLatSphericalPos]),
    (MathSphericalPos, type[MathSphericalPos]),
)
def represent_as(
    current: AbstractPos3D, target: type[AbstractPos3D], /, **kwargs: Any
) -> AbstractPos3D:
    """Self transforms for 3D vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    Cartesian to Cartesian:

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
    >>> cx.represent_as(vec, cx.CartesianPos3D) is vec
    True

    Cylindrical to Cylindrical:

    >>> vec = cx.CylindricalPos(rho=u.Quantity(1, "kpc"), phi=u.Quantity(2, "deg"),
    ...                         z=u.Quantity(3, "kpc"))
    >>> cx.represent_as(vec, cx.CylindricalPos) is vec
    True

    Spherical to Spherical:

    >>> vec = cx.SphericalPos(r=u.Quantity(1, "kpc"), theta=u.Quantity(2, "deg"),
    ...                       phi=u.Quantity(3, "deg"))
    >>> cx.represent_as(vec, cx.SphericalPos) is vec
    True

    LonLatSpherical to LonLatSpherical:

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(1, "deg"), lat=u.Quantity(2, "deg"),
    ...                             distance=u.Quantity(3, "kpc"))
    >>> cx.represent_as(vec, cx.LonLatSphericalPos) is vec
    True

    MathSpherical to MathSpherical:

    >>> vec = cx.MathSphericalPos(r=u.Quantity(1, "kpc"), theta=u.Quantity(2, "deg"),
    ...                           phi=u.Quantity(3, "deg"))
    >>> cx.represent_as(vec, cx.MathSphericalPos) is vec
    True

    """
    return current


@dispatch.multi(
    (CartesianVel3D, type[CartesianVel3D], AbstractPos),
    (CylindricalVel, type[CylindricalVel], AbstractPos),
    (SphericalVel, type[SphericalVel], AbstractPos),
    (LonLatSphericalVel, type[LonLatSphericalVel], AbstractPos),
    (
        LonCosLatSphericalVel,
        type[LonCosLatSphericalVel],
        AbstractPos,
    ),
    (MathSphericalVel, type[MathSphericalVel], AbstractPos),
    (ProlateSpheroidalVel, type[ProlateSpheroidalVel], AbstractPos),
)
def represent_as(
    current: AbstractVel3D,
    target: type[AbstractVel3D],
    position: AbstractPos,
    /,
    **kwargs: Any,
) -> AbstractVel3D:
    """Self transforms for 3D velocity.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    For these transformations the position does not matter since the
    self-transform returns the velocity unchanged.

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "kpc")

    Cartesian to Cartesian velocity:

    >>> dif = cx.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> cx.represent_as(dif, cx.CartesianVel3D, vec) is dif
    True

    Cylindrical to Cylindrical velocity:

    >>> dif = cx.CylindricalVel(d_rho=u.Quantity(1, "km/s"),
    ...                         d_phi=u.Quantity(2, "mas/yr"),
    ...                         d_z=u.Quantity(3, "km/s"))
    >>> cx.represent_as(dif, cx.CylindricalVel, vec) is dif
    True

    Spherical to Spherical velocity:

    >>> dif = cx.SphericalVel(d_r=u.Quantity(1, "km/s"),
    ...                       d_theta=u.Quantity(2, "mas/yr"),
    ...                       d_phi=u.Quantity(3, "mas/yr"))
    >>> cx.represent_as(dif, cx.SphericalVel, vec) is dif
    True

    LonLatSpherical to LonLatSpherical velocity:

    >>> dif = cx.LonLatSphericalVel(d_lon=u.Quantity(1, "mas/yr"),
    ...                             d_lat=u.Quantity(2, "mas/yr"),
    ...                             d_distance=u.Quantity(3, "km/s"))
    >>> cx.represent_as(dif, cx.LonLatSphericalVel, vec) is dif
    True

    LonCosLatSpherical to LonCosLatSpherical velocity:

    >>> dif = cx.LonCosLatSphericalVel(d_lon_coslat=u.Quantity(1, "mas/yr"),
    ...                                d_lat=u.Quantity(2, "mas/yr"),
    ...                                d_distance=u.Quantity(3, "km/s"))
    >>> cx.represent_as(dif, cx.LonCosLatSphericalVel, vec) is dif
    True

    MathSpherical to MathSpherical velocity:

    >>> dif = cx.MathSphericalVel(d_r=u.Quantity(1, "km/s"),
    ...                           d_theta=u.Quantity(2, "mas/yr"),
    ...                           d_phi=u.Quantity(3, "mas/yr"))
    >>> cx.represent_as(dif, cx.MathSphericalVel, vec) is dif
    True

    """
    return current


# =============================================================================
# CartesianPos3D


@dispatch
def represent_as(
    current: CartesianPos3D, target: type[CylindricalPos], /, **kwargs: Any
) -> CylindricalPos:
    """CartesianPos3D -> CylindricalPos.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(cx.represent_as(vec, cx.CylindricalPos))
    <CylindricalPos (rho[km], phi[rad], z[km])
        [2.236 1.107 3.   ]>

    """
    rho = xp.sqrt(current.x**2 + current.y**2)
    phi = xp.atan2(current.y, current.x)
    return target(rho=rho, phi=phi, z=current.z)


@dispatch
def represent_as(
    current: CartesianPos3D, target: type[SphericalPos], /, **kwargs: Any
) -> SphericalPos:
    """CartesianPos3D -> SphericalPos.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(cx.represent_as(vec, cx.SphericalPos))
    <SphericalPos (r[km], theta[rad], phi[rad])
        [3.742 0.641 1.107]>

    """
    r = xp.sqrt(current.x**2 + current.y**2 + current.z**2)
    theta = xp.acos(current.z / r)
    phi = xp.atan2(current.y, current.x)
    return target(r=r, theta=theta, phi=phi)


@dispatch.multi(
    (CartesianPos3D, type[LonLatSphericalPos]),
    (CartesianPos3D, type[MathSphericalPos]),
)
def represent_as(
    current: CartesianPos3D,
    target: type[AbstractSphericalPos],
    /,
    **kwargs: Any,
) -> AbstractSphericalPos:
    """CartesianPos3D -> AbstractSphericalPos.

    Examples
    --------
    >>> import coordinax as cx

    >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "km")

    >>> print(cx.represent_as(vec, cx.LonLatSphericalPos))
    <LonLatSphericalPos (lon[rad], lat[deg], distance[km])
        [ 1.107 53.301  3.742]>

    >>> print(cx.represent_as(vec, cx.MathSphericalPos))
    <MathSphericalPos (r[km], theta[rad], phi[rad])
        [3.742 1.107 0.641]>

    """
    return represent_as(represent_as(current, SphericalPos), target)


# =============================================================================
# CylindricalPos


@dispatch
def represent_as(
    current: CylindricalPos, target: type[CartesianPos3D], /, **kwargs: Any
) -> CartesianPos3D:
    """CylindricalPos -> CartesianPos3D.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CylindricalPos(rho=u.Quantity(1., "kpc"), phi=u.Quantity(90, "deg"),
    ...                         z=u.Quantity(1, "kpc"))
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-4.371e-08  1.000e+00  1.000e+00]>

    """
    x = current.rho * xp.cos(current.phi)
    y = current.rho * xp.sin(current.phi)
    z = current.z
    return target(x=x, y=y, z=z)


@dispatch
def represent_as(
    current: CylindricalPos, target: type[SphericalPos], /, **kwargs: Any
) -> SphericalPos:
    """CylindricalPos -> SphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CylindricalPos(rho=u.Quantity(1., "kpc"), phi=u.Quantity(90, "deg"),
    ...                         z=u.Quantity(1, "kpc"))
    >>> print(cx.represent_as(vec, cx.SphericalPos))
    <SphericalPos (r[kpc], theta[rad], phi[deg])
        [ 1.414  0.785 90.   ]>

    """
    r = xp.sqrt(current.rho**2 + current.z**2)
    theta = xp.acos(current.z / r)
    return target(r=r, theta=theta, phi=current.phi)


@dispatch.multi(
    (CylindricalPos, type[LonLatSphericalPos]),
    (CylindricalPos, type[MathSphericalPos]),
)
def represent_as(
    current: CylindricalPos,
    target: type[AbstractSphericalPos],
    /,
    **kwargs: Any,
) -> AbstractSphericalPos:
    """CylindricalPos -> AbstractSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.CylindricalPos(rho=u.Quantity(1., "kpc"), phi=u.Quantity(90, "deg"),
    ...                         z=u.Quantity(1, "kpc"))

    >>> print(cx.represent_as(vec, cx.LonLatSphericalPos))
    <LonLatSphericalPos (lon[deg], lat[deg], distance[kpc])
        [90.    45.     1.414]>

    >>> print(cx.represent_as(vec, cx.MathSphericalPos))
    <MathSphericalPos (r[kpc], theta[deg], phi[rad])
        [ 1.414 90.     0.785]>

    """
    return represent_as(represent_as(current, SphericalPos), target)


# =============================================================================
# SphericalPos


@dispatch
def represent_as(
    current: SphericalPos, target: type[CartesianPos3D], /, **kwargs: Any
) -> CartesianPos3D:
    """SphericalPos -> CartesianPos3D.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.SphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                       phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-4.371e-08  1.000e+00 -4.371e-08]>

    """
    x = current.r.distance * xp.sin(current.theta) * xp.cos(current.phi)
    y = current.r.distance * xp.sin(current.theta) * xp.sin(current.phi)
    z = current.r.distance * xp.cos(current.theta)
    return target(x=x, y=y, z=z)


@dispatch
def represent_as(
    current: SphericalPos, target: type[CylindricalPos], /, **kwargs: Any
) -> CylindricalPos:
    """SphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.SphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                       phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.CylindricalPos))
    <CylindricalPos (rho[kpc], phi[deg], z[kpc])
        [ 1.000e+00  9.000e+01 -4.371e-08]>

    """
    rho = xp.abs(current.r.distance * xp.sin(current.theta))
    z = current.r.distance * xp.cos(current.theta)
    return target(rho=rho, phi=current.phi, z=z)


@dispatch
def represent_as(
    current: SphericalPos, target: type[LonLatSphericalPos], /, **kwargs: Any
) -> LonLatSphericalPos:
    """SphericalPos -> LonLatSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.SphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                       phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.LonLatSphericalPos))
    <LonLatSphericalPos (lon[deg], lat[deg], distance[kpc])
        [90.  0.  1.]>

    """
    return target(
        lon=current.phi, lat=u.Quantity(90, "deg") - current.theta, distance=current.r
    )


@dispatch
def represent_as(
    current: SphericalPos, target: type[MathSphericalPos], /, **kwargs: Any
) -> MathSphericalPos:
    """SphericalPos -> MathSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.SphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                       phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.MathSphericalPos))
    <MathSphericalPos (r[kpc], theta[deg], phi[deg])
        [ 1. 90. 90.]>

    """
    return target(r=current.r, theta=current.phi, phi=current.theta)


# =============================================================================
# LonLatSphericalPos


@dispatch
def represent_as(
    current: LonLatSphericalPos,
    target: type[CartesianPos3D],
    /,
    **kwargs: Any,
) -> CartesianPos3D:
    """LonLatSphericalPos -> CartesianPos3D.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(90, "deg"),
    ...                             lat=u.Quantity(0, "deg"),
    ...                             distance=u.Quantity(1., "kpc"))
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-4.371e-08  1.000e+00 -4.371e-08]>

    """
    return represent_as(represent_as(current, SphericalPos), CartesianPos3D)


@dispatch
def represent_as(
    current: LonLatSphericalPos,
    target: type[CylindricalPos],
    /,
    **kwargs: Any,
) -> CylindricalPos:
    """LonLatSphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(90, "deg"),
    ...                             lat=u.Quantity(0, "deg"),
    ...                             distance=u.Quantity(1., "kpc"))
    >>> print(cx.represent_as(vec, cx.CylindricalPos))
    <CylindricalPos (rho[kpc], phi[deg], z[kpc])
        [ 1.000e+00  9.000e+01 -4.371e-08]>

    """
    return represent_as(represent_as(current, SphericalPos), target)


@dispatch
def represent_as(
    current: LonLatSphericalPos, target: type[SphericalPos], /, **kwargs: Any
) -> SphericalPos:
    """LonLatSphericalPos -> SphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(90, "deg"),
    ...                             lat=u.Quantity(0, "deg"),
    ...                             distance=u.Quantity(1., "kpc"))
    >>> print(cx.represent_as(vec, cx.SphericalPos))
    <SphericalPos (r[kpc], theta[deg], phi[deg])
        [ 1. 90. 90.]>

    """
    return target(
        r=current.distance, theta=u.Quantity(90, "deg") - current.lat, phi=current.lon
    )


# =============================================================================
# MathSphericalPos


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[CartesianPos3D], /, **kwargs: Any
) -> CartesianPos3D:
    """MathSphericalPos -> CartesianPos3D.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.MathSphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                           phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-4.371e-08  1.000e+00 -4.371e-08]>

    """
    x = current.r.distance * xp.sin(current.phi) * xp.cos(current.theta)
    y = current.r.distance * xp.sin(current.phi) * xp.sin(current.theta)
    z = current.r.distance * xp.cos(current.phi)
    return target(x=x, y=y, z=z)


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[CylindricalPos], /, **kwargs: Any
) -> CylindricalPos:
    """MathSphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.MathSphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                           phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.CylindricalPos))
    <CylindricalPos (rho[kpc], phi[deg], z[kpc])
        [ 1.000e+00  9.000e+01 -4.371e-08]>

    """
    rho = xp.abs(current.r.distance * xp.sin(current.phi))
    z = current.r.distance * xp.cos(current.phi)
    return target(rho=rho, phi=current.theta, z=z)


@dispatch
def represent_as(
    current: MathSphericalPos, target: type[SphericalPos], /, **kwargs: Any
) -> SphericalPos:
    """MathSphericalPos -> SphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.MathSphericalPos(r=u.Quantity(1., "kpc"), theta=u.Quantity(90, "deg"),
    ...                           phi=u.Quantity(90, "deg"))
    >>> print(cx.represent_as(vec, cx.SphericalPos))
    <SphericalPos (r[kpc], theta[deg], phi[deg])
        [ 1. 90. 90.]>

    """
    return target(r=current.r, theta=current.phi, phi=current.theta)


# =============================================================================
# ProlateSpheroidalPos


@dispatch
def represent_as(
    current: ProlateSpheroidalPos, target: type[CylindricalPos], /, **kwargs: Any
) -> CylindricalPos:
    """ProlateSpheroidalPos -> CylindricalPos.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(1., "kpc2"),
    ...     nu=Quantity(0.2, "kpc2"),
    ...     phi=Quantity(90, "deg"),
    ...     Delta=Quantity(0.5, "kpc")
    ... )
    >>> print(cx.represent_as(vec, cx.CylindricalPos))
    <CylindricalPos (rho[kpc], phi[deg], z[kpc])
        [ 0.387 90.     0.894]>

    """
    Delta2 = current.Delta**2
    nu_D2 = xp.abs(current.nu) / Delta2

    R = xp.sqrt((current.mu - Delta2) * (1 - nu_D2))
    z = xp.sqrt(current.mu * nu_D2) * xp.sign(current.nu)

    return target(rho=R, phi=current.phi, z=z)


@dispatch
def represent_as(
    current: CylindricalPos,
    target: type[ProlateSpheroidalPos],
    /,
    **kwargs: Any,
) -> ProlateSpheroidalPos:
    """CylindricalPos -> ProlateSpheroidalPos.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CylindricalPos(
    ...     rho=Quantity(1., "kpc"),
    ...     phi=Quantity(90, "deg"),
    ...     z=Quantity(1, "kpc")
    ... )
    >>> print(vec.represent_as(cx.ProlateSpheroidalPos, Delta=Quantity(0.5, "kpc")))
    <ProlateSpheroidalPos (mu[kpc2], nu[kpc2], phi[deg])
        [ 2.133  0.117 90.   ]>

    """
    Delta = eqx.error_if(
        kwargs.get("Delta"),
        "Delta" not in kwargs,
        "Delta must be provided for ProlateSpheroidalPos.",
    )
    R2 = current.rho**2
    z2 = current.z**2
    Delta2 = Delta**2

    sum_ = R2 + z2 + Delta2
    diff_ = R2 + z2 - Delta2

    # compute D = sqrt((R² + z² - Δ²)² + 4R²Δ²)
    D = xp.sqrt(diff_**2 + 4 * R2 * Delta2)

    # handle special cases for R=0 or z=0
    # TODO: quaxed.numpy.select doesn't work with Quantity's?
    # D = xp.select(
    #     [current.z == 0, current.rho == 0, xp.full(current.rho.shape, 1, dtype=bool)],
    #     [
    #         sum_,  # z=0 case
    #         xp.abs(diff_),  # R=0 case
    #         D,  # otherwise
    #     ],
    # )
    D = xp.where(current.z == 0, sum_, D)
    D = xp.where(current.rho == 0, xp.abs(diff_), D)

    # compute mu and nu depending on sign of diff_ - avoids dividing by a small number
    pos_mu_minus_delta = 0.5 * (D + diff_)
    pos_delta_minus_nu = Delta2 * R2 / pos_mu_minus_delta

    neg_delta_minus_nu = 0.5 * (D - diff_)
    neg_mu_minus_delta = Delta2 * R2 / neg_delta_minus_nu

    # Select based on condition
    mu_minus_delta = xp.where(diff_ >= 0, pos_mu_minus_delta, neg_mu_minus_delta)
    delta_minus_nu = xp.where(diff_ >= 0, pos_delta_minus_nu, neg_delta_minus_nu)

    # compute mu and nu:
    mu = Delta2 + mu_minus_delta
    abs_nu = 2 * Delta2 / (sum_ + D) * z2

    # for numerical stability when Delta^2-|nu| is small
    abs_nu = xp.where(abs_nu * 2 > Delta2, Delta2 - delta_minus_nu, abs_nu)

    nu = abs_nu * xp.sign(current.z)

    return target(mu=mu, nu=nu, phi=current.phi, Delta=Delta)


@dispatch
def represent_as(
    current: ProlateSpheroidalPos, target: type[AbstractPos3D], /, **kwargs: Any
) -> AbstractPos3D:
    """ProlateSpheroidalPos -> AbstractPos3D.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(1., "kpc2"),
    ...     nu=Quantity(0.2, "kpc2"),
    ...     phi=Quantity(90, "deg"),
    ...     Delta=Quantity(0.5, "kpc")
    ... )
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-1.693e-08  3.873e-01  8.944e-01]>

    """
    cyl = represent_as(current, CylindricalPos)
    return represent_as(cyl, target)


@dispatch
def represent_as(
    current: ProlateSpheroidalPos,
    target: type[ProlateSpheroidalPos],
    /,
    **kwargs: Any,
) -> ProlateSpheroidalPos:
    """ProlateSpheroidalPos -> ProlateSpheroidalPos.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    Self-transforms can change the focal length:

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(1., "kpc2"),
    ...     nu=Quantity(0.2, "kpc2"),
    ...     phi=Quantity(90, "deg"),
    ...     Delta=Quantity(0.5, "kpc")
    ... )
    >>> print(cx.represent_as(vec, cx.ProlateSpheroidalPos, Delta=Quantity(0.8, "kpc")))
    <ProlateSpheroidalPos...>

    Without changing the focal length, no transform is done:

    >>> vec2 = cx.represent_as(vec, cx.ProlateSpheroidalPos)
    >>> vec == vec2
    Array(True, dtype=bool)

    """
    Delta = kwargs.get("Delta", current.Delta)
    return jax.lax.cond(
        "Delta" in kwargs,
        lambda Delta: represent_as(
            represent_as(current, CylindricalPos), target, Delta=Delta
        ),
        lambda _: current,
        Delta,
    )


@dispatch
def represent_as(
    current: AbstractPos3D,
    target: type[ProlateSpheroidalPos],
    /,
    **kwargs: Any,
) -> ProlateSpheroidalPos:
    """AbstractPos3D -> ProlateSpheroidalPos.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(1., "kpc2"),
    ...     nu=Quantity(0.2, "kpc2"),
    ...     phi=Quantity(90, "deg"),
    ...     Delta=Quantity(0.5, "kpc")
    ... )
    >>> print(cx.represent_as(vec, cx.CartesianPos3D))
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [-1.693e-08  3.873e-01  8.944e-01]>

    Self-transforms also work to change the focal length:

    >>> vec = cx.ProlateSpheroidalPos(
    ...     mu=Quantity(1., "kpc2"),
    ...     nu=Quantity(0.2, "kpc2"),
    ...     phi=Quantity(90, "deg"),
    ...     Delta=Quantity(0.5, "kpc")
    ... )
    >>> print(cx.represent_as(vec, cx.ProlateSpheroidalPos, Delta=Quantity(0.8, "kpc")))
    <ProlateSpheroidalPos...>

    """
    Delta = eqx.error_if(
        kwargs.get("Delta"),
        "Delta" not in kwargs,
        "Delta must be provided for ProlateSpheroidalPos.",
    )
    cyl = represent_as(current, CylindricalPos)
    return represent_as(cyl, target, Delta=Delta)


# =============================================================================
# LonLatSphericalVel


@dispatch
def represent_as(
    current: AbstractVel3D,
    target: type[LonCosLatSphericalVel],
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> LonCosLatSphericalVel:
    """AbstractVel3D -> LonCosLatSphericalVel.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> vec = cx.LonLatSphericalPos(lon=u.Quantity(15, "deg"),
    ...                             lat=u.Quantity(10, "deg"),
    ...                             distance=u.Quantity(1.5, "kpc"))
    >>> dif = cx.LonLatSphericalVel(d_lon=u.Quantity(7, "mas/yr"),
    ...                             d_lat=u.Quantity(0, "deg/Gyr"),
    ...                             d_distance=u.Quantity(-5, "km/s"))
    >>> newdif = cx.represent_as(dif, cx.LonCosLatSphericalVel, vec)
    >>> newdif
    LonCosLatSphericalVel(
      d_lon_coslat=Quantity[...]( value=f32[], unit=Unit("mas / yr") ),
      d_lat=Quantity[...]( value=f32[], unit=Unit("deg / Gyr") ),
      d_distance=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    >>> newdif.d_lon_coslat / jnp.cos(vec.lat)  # float32 imprecision
    Quantity['angular frequency'](Array(6.9999995, dtype=float32), unit='mas / yr')

    """
    # Parse the position to an AbstractPos
    if isinstance(position, AbstractPos):
        posvec = position
    else:  # Q -> Cart<X>D
        posvec = current.integral_cls._cartesian_cls.from_(  # noqa: SLF001
            position
        )

    # Transform the differential to LonLatSphericalVel
    current = represent_as(current, LonLatSphericalVel, posvec)

    # Transform the position to the required type
    posvec = represent_as(posvec, current.integral_cls)

    # Calculate the differential in the new system
    return target(
        d_lon_coslat=current.d_lon * xp.cos(posvec.lat),
        d_lat=current.d_lat,
        d_distance=current.d_distance,
    )


@dispatch
def represent_as(
    current: LonCosLatSphericalVel,
    target: type[LonLatSphericalVel],
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> LonLatSphericalVel:
    """LonCosLatSphericalVel -> LonLatSphericalVel."""
    # Parse the position to an AbstractPos
    if isinstance(position, AbstractPos):
        posvec = position
    else:  # Q -> Cart<X>D
        posvec = current.integral_cls._cartesian_cls.from_(  # noqa: SLF001
            position
        )

    # Transform the position to the required type
    posvec = represent_as(posvec, current.integral_cls)

    # Calculate the differential in the new system
    return target(
        d_lon=current.d_lon_coslat / xp.cos(posvec.lat),
        d_lat=current.d_lat,
        d_distance=current.d_distance,
    )


@dispatch
def represent_as(
    current: LonCosLatSphericalVel,
    target: type[AbstractVel3D],
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> AbstractVel3D:
    """LonCosLatSphericalVel -> AbstractVel3D."""
    # Parse the position to an AbstractPos
    if isinstance(position, AbstractPos):
        posvec = position
    else:  # Q -> Cart<X>D
        posvec = current.integral_cls._cartesian_cls.from_(  # noqa: SLF001
            position
        )
    # Transform the differential to LonLatSphericalVel
    current = represent_as(current, LonLatSphericalVel, posvec)
    # Transform the position to the required type
    return represent_as(current, target, posvec)


# =============================================================================
# CartesianVel3D


@dispatch
def represent_as(
    current: CartesianVel3D, target: type[CartesianVel3D], /
) -> CartesianVel3D:
    """CartesianVel3D -> CartesianVel3D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> v = cx.CartesianVel3D.from_([1, 1, 1], "m/s")
    >>> cx.represent_as(v, cx.CartesianVel3D) is v
    True

    """
    return current


# =============================================================================
# CartesianAcc3D


@dispatch
def represent_as(
    current: CartesianAcc3D, target: type[CartesianAcc3D], /
) -> CartesianAcc3D:
    """CartesianAcc3D -> CartesianAcc3D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian vectors.

    Examples
    --------
    >>> import coordinax as cx
    >>> a = cx.CartesianAcc3D.from_([1, 1, 1], "m/s2")
    >>> cx.represent_as(a, cx.CartesianAcc3D) is a
    True

    """
    return current



"""3-dimensional representations."""
# pylint: disable=duplicate-code

__all__ = [
    # Base
    "AbstractPos3D",
    "AbstractVel3D",
    "AbstractAcc3D",
    # Cartesian
    "CartesianPos3D",
    "CartesianVel3D",
    "CartesianAcc3D",
    # Cylindrical
    "CylindricalPos",
    "CylindricalVel",
    "CylindricalAcc",
    # Base Spherical
    "AbstractSphericalPos",
    "AbstractSphericalVel",
    "AbstractSphericalAcc",
    # Spherical
    "SphericalPos",
    "SphericalVel",
    "SphericalAcc",
    # Math Spherical
    "MathSphericalPos",
    "MathSphericalVel",
    "MathSphericalAcc",
    # LonLat Spherical
    "LonLatSphericalPos",
    "LonLatSphericalVel",
    "LonLatSphericalAcc",
    "LonCosLatSphericalVel",
    # Prolate Spheroidal
    "ProlateSpheroidalPos",
    "ProlateSpheroidalVel",
    "ProlateSpheroidalAcc",
    # Generic
    "CartesianGeneric3D",
]

from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from .base_spherical import (
    AbstractSphericalAcc,
    AbstractSphericalPos,
    AbstractSphericalVel,
)
from .cartesian import CartesianAcc3D, CartesianPos3D, CartesianVel3D
from .cylindrical import CylindricalAcc, CylindricalPos, CylindricalVel
from .generic import CartesianGeneric3D
from .lonlatspherical import (
    LonCosLatSphericalVel,
    LonLatSphericalAcc,
    LonLatSphericalPos,
    LonLatSphericalVel,
)
from .mathspherical import MathSphericalAcc, MathSphericalPos, MathSphericalVel
from .spherical import SphericalAcc, SphericalPos, SphericalVel
from .spheroidal import ProlateSpheroidalAcc, ProlateSpheroidalPos, ProlateSpheroidalVel

# isort: split
from . import (
    compat,  # noqa: F401
    constructor,  # noqa: F401
    transform,  # noqa: F401
)



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanSpatialTranslationOperator", "GalileanTranslationOperator"]


from dataclasses import replace
from typing import Any, Literal, final

import equinox as eqx
from plum import convert, dispatch

import quaxed.numpy as jnp
import unxt as u

from .base import AbstractGalileanOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.identity import IdentityOperator
from coordinax._src.vectors.base import AbstractPos
from coordinax._src.vectors.d1 import CartesianPos1D
from coordinax._src.vectors.d2 import CartesianPos2D
from coordinax._src.vectors.d3 import AbstractPos3D, CartesianPos3D
from coordinax._src.vectors.d4 import FourVector

##############################################################################
# Spatial Translations


def _converter_spatialtranslation(x: Any) -> AbstractPos:
    """Convert to a spatial translation vector."""
    out: AbstractPos | None = None
    if isinstance(x, GalileanSpatialTranslationOperator):
        out = x.translation
    elif isinstance(x, AbstractPos):
        out = x
    elif isinstance(x, u.Quantity):
        shape: tuple[int, ...] = x.shape
        match shape:
            case (1,):
                out = CartesianPos1D.from_(x)
            case (2,):
                out = CartesianPos2D.from_(x)
            case (3,):
                out = CartesianPos3D.from_(x)
            case _:
                msg = f"Cannot convert {x} to a spatial translation vector."
                raise TypeError(msg)

    if out is None:
        msg = f"Cannot convert {x} to a spatial translation vector."
        raise TypeError(msg)

    return out


@final
class GalileanSpatialTranslationOperator(AbstractGalileanOperator):
    r"""Operator for Galilean spatial translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    Parameters
    ----------
    translation : :class:`vector.AbstractPos3D`
        The spatial translation vector. This parameters accepts either a
        :class:`vector.AbstractPos3D` instance or uses
        :meth:`vector.CartesianPos3D.from_` to enable a variety of more
        convenient input types to create a Cartesian vector. See
        :class:`vector.CartesianPos3D` for details.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import coordinax.operators as cxo

    We can then create a spatial translation operator:

    >>> shift = u.Quantity([1.0, 2.0, 3.0], "kpc")
    >>> op = cxo.GalileanSpatialTranslationOperator(shift)
    >>> op
    GalileanSpatialTranslationOperator( translation=CartesianPos3D( ... ) )

    Note that the translation is a :class:`vector.CartesianPos3D`, which was
    constructed from a 1D array, using :meth:`vector.CartesianPos3D.from_`. We
    can also construct it directly, which allows for other vector types.

    >>> shift = cx.SphericalPos(r=u.Quantity(1.0, "kpc"),
    ...                         theta=u.Quantity(jnp.pi/2, "rad"),
    ...                         phi=u.Quantity(0, "rad"))
    >>> op = cxo.GalileanSpatialTranslationOperator(shift)
    >>> op
    GalileanSpatialTranslationOperator( translation=SphericalPos( ... ) )

    Translation operators can be applied to :class:`vector.AbstractPos`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "kpc")
    >>> op(q)
    CartesianPos3D( ... )

    And to :class:`~unxt.Quantity`:

    >>> q = u.Quantity([0, 0, 0], "kpc")
    >>> op(q).value.round(2)
    Array([ 1.,  0., -0.], dtype=float32)

    :class:`coordinax.operators.GalileanSpatialTranslationOperator` can be used
    for other dimensional vectors as well:

    - 1D:

    >>> op = cxo.GalileanSpatialTranslationOperator.from_([1], "kpc")
    >>> q = u.Quantity([0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPos1D.from_(q).represent_as(cx.RadialPos)
    >>> op(vec)
    RadialPos(r=Distance(value=f32[], unit=Unit("kpc")))

    - 2D:

    >>> op = cxo.GalileanSpatialTranslationOperator.from_([1, 2], "kpc")
    >>> q = u.Quantity([0, 0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1., 2.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPos2D.from_(q).represent_as(cx.PolarPos)
    >>> op(vec)
    PolarPos( r=Distance(value=f32[], unit=Unit("kpc")),
              phi=Angle(value=f32[], unit=Unit("rad")) )

    - 3D:

    >>> op = cxo.GalileanSpatialTranslationOperator.from_([1, 2, 3], "kpc")
    >>> q = u.Quantity([0, 0, 0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1., 2., 3.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPos3D.from_(q).represent_as(cx.SphericalPos)
    >>> op(vec)
    SphericalPos( r=Distance(value=f32[], unit=Unit("kpc")),
                  theta=Angle(value=f32[], unit=Unit("rad")),
                  phi=Angle(value=f32[], unit=Unit("rad")) )

    Many operators are time dependent and require a time argument. This operator
    is time independent and will pass through the time argument:

    >>> t = u.Quantity(0, "Gyr")
    >>> op(q, t)[1] is t
    True

    """

    translation: AbstractPos = eqx.field(converter=_converter_spatialtranslation)
    """The spatial translation.

    This parameters accepts either a :class:`vector.AbstractVector` instance or
    uses a Cartesian vector from_ to enable a variety of more convenient
    input types to create a Cartesian vector. See
    :class:`vector.CartesianPos3D.from_` for an example when doing a 3D
    translation.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial frame-preserving transformation.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = GalileanSpatialTranslationOperator.from_([1, 1, 1], "kpc")

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanSpatialTranslationOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> import coordinax.operators as cxo

        >>> op = cxo.GalileanSpatialTranslationOperator.from_([1, 1, 1], "kpc")

        >>> op.inverse
        GalileanSpatialTranslationOperator( translation=CartesianPos3D( ... ) )

        >>> print(op.inverse.translation)
        <CartesianPos3D (x[kpc], y[kpc], z[kpc])
            [-1. -1. -1.]>

        """
        return GalileanSpatialTranslationOperator(-self.translation)

    # -------------------------------------------

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator", q: AbstractPos, /
    ) -> AbstractPos:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        >>> shift = cx.CartesianPos3D.from_([1, 1, 1], "kpc")
        >>> op = cxo.GalileanSpatialTranslationOperator(shift)

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> t = u.Quantity(0, "Gyr")
        >>> newq = op(q)
        >>> print(newq)
        <CartesianPos3D (x[kpc], y[kpc], z[kpc])
            [2. 3. 4.]>

        """
        return q + self.translation

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator",
        q: AbstractPos,
        t: u.Quantity["time"],
        /,
    ) -> tuple[AbstractPos, u.Quantity["time"]]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        >>> shift = cx.CartesianPos3D.from_([1, 1, 1], "kpc")
        >>> op = cxo.GalileanSpatialTranslationOperator(shift)

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> t = u.Quantity(0, "Gyr")
        >>> newq, newt = op(q, t)
        >>> print(newq)
        <CartesianPos3D (x[kpc], y[kpc], z[kpc])
            [2. 3. 4.]>

        This spatial translation is time independent.

        >>> op(q, u.Quantity(1, "Gyr"))[0].x == newq.x
        Array(True, dtype=bool)

        """
        return q + self.translation, t

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator", v4: FourVector, /
    ) -> AbstractPos:
        """Apply the translation to the coordinates."""  # TODO: docstring
        return replace(v4, q=v4.q + self.translation)


@dispatch  # type: ignore[misc]
def simplify_op(
    op: GalileanSpatialTranslationOperator, /, **kwargs: Any
) -> AbstractOperator:
    """Simplify a Galilean spatial translation operator.

    Examples
    --------
    >>> import coordinax.operators as co

    An operator with real effect cannot be simplified:

    >>> op = co.GalileanSpatialTranslationOperator.from_([1, 0, 0], "m")
    >>> co.simplify_op(op)
    GalileanSpatialTranslationOperator(
      translation=CartesianPos3D( ... )
    )

    An operator with no effect can be simplified:

    >>> op = co.GalileanSpatialTranslationOperator.from_([0, 0, 0], "m")
    >>> co.simplify_op(op)
    IdentityOperator()

    """
    # Check if the translation is zero.
    if jnp.allclose(
        convert(op.translation, u.Quantity).value, jnp.zeros((3,)), **kwargs
    ):
        return IdentityOperator()
    return op


##############################################################################


@final
class GalileanTranslationOperator(AbstractGalileanOperator):
    r"""Operator for spatio-temporal translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    where :math:`a \in R^3` and :math:`s \in R`.  Therefore for a potential
    :math:`\Phi(t,\mathbf{x})` in the translated frame the potential is given by
    the subtraction of the translation.

    Parameters
    ----------
    translation : :class:`vector.FourVector`
        The translation vector [T, Q].  This parameters uses
        :meth:`vector.FourVector.from_` to enable a variety of more
        convenient input types. See :class:`vector.FourVector` for details.

    Examples
    --------
    We start with the required imports:

    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx
    >>> import coordinax.operators as cxo

    We can then create a translation operator:

    >>> op = cxo.GalileanTranslationOperator.from_([1.0, 2.0, 3.0, 4.0], "kpc")
    >>> op
    GalileanTranslationOperator(
      translation=FourVector(
        t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
        q=CartesianPos3D( ... ) )
    )

    Note that the translation is a :class:`vector.FourVector`, which was
    constructed from a 1D array, using :meth:`vector.FourVector.from_`. We
    can also construct it directly, which allows for other vector types.

    >>> qshift = cx.SphericalPos(r=u.Quantity(1.0, "kpc"),
    ...                          theta=u.Quantity(jnp.pi/2, "rad"),
    ...                          phi=u.Quantity(0, "rad"))
    >>> shift = cx.FourVector(u.Quantity(1.0, "Gyr"), qshift)
    >>> op = cxo.GalileanTranslationOperator(shift)
    >>> op
    GalileanTranslationOperator(
      translation=FourVector(
        t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("Gyr")),
        q=SphericalPos( ... ) )
    )

    Translation operators can be applied to :class:`vector.FourVector`:

    >>> w = cx.FourVector.from_([0, 0, 0, 0], "kpc")
    >>> op(w)
    FourVector(
      t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
      q=CartesianPos3D( ... )
    )

    Also to :class:`vector.AbstractPos3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPos3D.from_([0, 0, 0], "kpc")
    >>> t = u.Quantity(0, "Gyr")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(1., dtype=float32), unit='kpc')
    >>> newt
    Quantity['time'](Array(1., dtype=float32), unit='Gyr')

    """

    translation: FourVector = eqx.field(converter=FourVector.from_)
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters uses
    :meth:`vector.FourVector.from_` to enable a variety of more convenient
    input types. See :class:`vector.FourVector` for details.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial-frame preserving transformation.

        Examples
        --------
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        >>> shift = cx.FourVector.from_([0, 1, 1, 1], "kpc")
        >>> op = cxo.GalileanTranslationOperator(shift)

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanTranslationOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "kpc")
        >>> shift = FourVector(u.Quantity(1, "Gyr"), qshift)
        >>> op = cxo.GalileanTranslationOperator(shift)

        >>> op.inverse
        GalileanTranslationOperator( translation=FourVector( ... ) )

        >>> op.inverse.translation.q.x
        Quantity['length'](Array(-1., dtype=float32), unit='kpc')

        """
        return GalileanTranslationOperator(-self.translation)

    # -------------------------------------------

    @AbstractOperator.__call__.dispatch
    def __call__(self: "GalileanTranslationOperator", x: FourVector, /) -> FourVector:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "kpc")
        >>> shift = FourVector(u.Quantity(1, "Gyr"), qshift)
        >>> op = cxo.GalileanTranslationOperator(shift)

        Construct a vector to translate, using the convenience from_ (the
        0th component is :math:`c * t`, the rest are spatial components):

        >>> w = cx.FourVector.from_([0, 1, 2, 3], "kpc")
        >>> w.t
        Quantity['time'](Array(0., dtype=float32), unit='kpc s / km')

        Apply the translation operator:

        >>> new = op(w)
        >>> new.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        >>> new.t.uconvert("Gyr")
        Quantity['time'](Array(0.99999994, dtype=float32), unit='Gyr')

        """
        return x + self.translation

    @AbstractOperator.__call__.dispatch(precedence=1)
    def __call__(
        self: "GalileanTranslationOperator",
        x: AbstractPos3D,
        t: u.Quantity["time"],
        /,
    ) -> tuple[AbstractPos3D, u.Quantity["time"]]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx
        >>> import coordinax.operators as cxo

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPos3D.from_([1, 1, 1], "kpc")
        >>> tshift = u.Quantity(1, "Gyr")
        >>> shift = cx.FourVector(tshift, qshift)
        >>> op = cxo.GalileanTranslationOperator(shift)

        Construct a vector to translate

        >>> q = cx.CartesianPos3D.from_([1, 2, 3], "kpc")
        >>> t = u.Quantity(1, "Gyr")
        >>> newq, newt = op(q, t)

        >>> newq.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        >>> newt
        Quantity['time'](Array(2., dtype=float32), unit='Gyr')

        """
        return (x + self.translation.q, t + self.translation.t)


@dispatch  # type: ignore[misc]
def simplify_op(op: GalileanTranslationOperator, /, **kwargs: Any) -> AbstractOperator:
    """Simplify a Galilean translation operator.

    Examples
    --------
    >>> import coordinax.operators as co

    An operator with real effect cannot be simplified:

    >>> op = co.GalileanTranslationOperator.from_([3e8, 1, 0, 0], "m")
    >>> co.simplify_op(op)
    GalileanTranslationOperator(
      translation=FourVector( ... )
    )

    An operator with no effect can be simplified:

    >>> op = co.GalileanTranslationOperator.from_([0, 0, 0, 0], "m")
    >>> co.simplify_op(op)
    IdentityOperator()

    """
    # Check if the translation is zero.
    if jnp.allclose(
        convert(op.translation, u.Quantity).value, jnp.zeros((4,)), **kwargs
    ):
        return IdentityOperator()
    # TODO: Check if the translation is purely spatial.

    return op



"""Base implementation of coordinate frames."""

__all__ = ["NoFrame"]


from typing import final

from .base import AbstractReferenceFrame


@final
class NoFrame(AbstractReferenceFrame):
    """A null reference frame.

    This is a reference frame that cannot be transformed to or from.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> null = cxf.NoFrame()
    >>> icrs = cxf.ICRS()

    >>> try:
    ...     cxf.frame_transform_op(null, icrs)
    ... except cxf.FrameTransformError as e:
    ...     print(e)
    Cannot transform from the null frame.

    >>> try:
    ...     cxf.frame_transform_op(icrs, null)
    ... except cxf.FrameTransformError as e:
    ...     print(e)
    Cannot transform to the null frame.

    """



"""Compatibility for Quantity."""

__all__: list[str] = []

from plum import add_promotion_rule, conversion_method

from unxt.quantity import AbstractQuantity, Quantity

from .base import AbstractDistance
from .core import Distance, DistanceModulus, Parallax

# Add a rule that when a AbstractDistance interacts with a Quantity, the
# distance degrades to a Quantity. This is necessary for many operations, e.g.
# division of a distance by non-dimensionless quantity where the resulting units
# are not those of a distance.
add_promotion_rule(AbstractDistance, Quantity, Quantity)

#####################################################################
# Conversion


@conversion_method(type_from=AbstractDistance, type_to=Quantity)  # type: ignore[misc]
def _convert_distance_to_quantity(x: AbstractDistance) -> Quantity:
    """Convert a distance to a quantity."""
    return Quantity(x.value, x.unit)


@conversion_method(type_from=AbstractQuantity, type_to=Distance)  # type: ignore[misc]
def _quantity_to_distance(q: AbstractQuantity, /) -> Distance:
    """Convert any quantity to a Distance.

    Examples
    --------
    >>> from plum import convert
    >>> from unxt.quantity import Quantity, UncheckedQuantity
    >>> from coordinax.distance import Distance
    >>> q = UncheckedQuantity(1, "m")
    >>> q
    UncheckedQuantity(Array(1, dtype=int32, ...), unit='m')

    >>> convert(q, Distance)
    Distance(Array(1, dtype=int32, ...), unit='m')

    The self-conversion doesn't copy the object:

    >>> q = Distance(1, "m")
    >>> convert(q, Distance) is q
    True

    """
    if isinstance(q, Distance):
        return q
    return Distance(q.value, q.unit)


@conversion_method(type_from=AbstractQuantity, type_to=Parallax)  # type: ignore[misc]
def _quantity_to_parallax(q: AbstractQuantity, /) -> Parallax:
    """Convert any quantity to a Parallax.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> from coordinax.distance import Parallax
    >>> q = u.Quantity(1, "mas")
    >>> q
    Quantity['angle'](Array(1, dtype=int32, ...), unit='mas')

    >>> convert(q, Parallax)
    Parallax(Array(1, dtype=int32, weak_type=True), unit='mas')

    The self-conversion doesn't copy the object:

    >>> q = Parallax(1, "mas")
    >>> convert(q, Parallax) is q
    True

    """
    if isinstance(q, Parallax):
        return q
    return Parallax(q.value, q.unit)


@conversion_method(type_from=AbstractQuantity, type_to=DistanceModulus)  # type: ignore[misc]
def _quantity_to_distmod(q: AbstractQuantity, /) -> DistanceModulus:
    """Convert any quantity to a DistanceModulus.

    Examples
    --------
    >>> from plum import convert
    >>> import unxt as u
    >>> from coordinax.distance import DistanceModulus
    >>> q = u.Quantity(1, "mag")
    >>> q
    Quantity['dex'](Array(1, dtype=int32, ...), unit='mag')

    >>> convert(q, DistanceModulus)
    DistanceModulus(Array(1, dtype=int32, ...), unit='mag')

    The self-conversion doesn't copy the object:

    >>> q = DistanceModulus(1, "mag")
    >>> convert(q, DistanceModulus) is q
    True

    """
    if isinstance(q, DistanceModulus):
        return q
    return DistanceModulus(q.value, q.unit)



"""Representation of velocities in different systems."""

__all__ = ["AbstractVel"]

from abc import abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Any, TypeVar

import equinox as eqx
import jax
from quax import register

import quaxed.numpy as jnp
import unxt as u
from dataclassish import field_items

from .base import AbstractVector
from .base_pos import AbstractPos
from coordinax._src.utils import classproperty

if TYPE_CHECKING:
    from typing import Self

VelT = TypeVar("VelT", bound="AbstractVel")

DIFFERENTIAL_CLASSES: set[type["AbstractVel"]] = set()


class AbstractVel(AbstractVector):  # pylint: disable=abstract-method
    """Abstract representation of vector differentials in different systems."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the subclass.

        The subclass is registered.
        """
        DIFFERENTIAL_CLASSES.add(cls)

    @classproperty
    @classmethod
    @abstractmethod
    def _cartesian_cls(cls) -> type["AbstractVector"]:
        """Return the corresponding Cartesian vector class.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.RadialVel._cartesian_cls
        <class 'coordinax...CartesianVel1D'>

        >>> cx.SphericalVel._cartesian_cls
        <class 'coordinax...CartesianVel3D'>

        """
        # TODO: something nicer than this for getting the corresponding class
        raise NotImplementedError

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type["AbstractPos"]:
        """Return the corresponding vector class.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.RadialVel.integral_cls.__name__
        'RadialPos'

        >>> cx.SphericalVel.integral_cls.__name__
        'SphericalPos'

        """
        raise NotImplementedError

    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> type["AbstractAcc"]:
        """Return the corresponding differential vector class.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.RadialVel.differential_cls.__name__
        'RadialAcc'

        >>> cx.SphericalVel.differential_cls.__name__
        'SphericalAcc'

        """
        raise NotImplementedError

    # ===============================================================
    # Quax

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        raise NotImplementedError

    # ===============================================================
    # Unary operations

    def __neg__(self) -> "Self":
        """Negate the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> dr = cx.RadialVel.from_([1], "m/s")
        >>> -dr
        RadialVel( d_r=Quantity[...]( value=i32[], unit=Unit("m / s") ) )

        >>> dp = cx.PolarVel(u.Quantity(1, "m/s"), u.Quantity(1, "mas/yr"))
        >>> neg_dp = -dp
        >>> neg_dp.d_r
        Quantity['speed'](Array(-1., dtype=float32), unit='m / s')
        >>> neg_dp.d_phi
        Quantity['angular frequency'](Array(-1., dtype=float32), unit='mas / yr')

        """
        return jax.tree.map(jnp.negative, self)

    # ===============================================================
    # Convenience methods

    @partial(eqx.filter_jit, inline=True)
    def norm(self, position: AbstractPos, /) -> u.Quantity["speed"]:
        """Return the norm of the vector."""
        return self.represent_as(self._cartesian_cls, position).norm()


# ---------------------------------------------------------


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_vel_q(self: AbstractVel, other: u.Quantity["time"]) -> AbstractPos:
    """Multiply the vector by a time :class:`unxt.Quantity` to get a position.

    Examples
    --------
    >>> from quaxed import lax
    >>> import unxt as u
    >>> import coordinax as cx

    >>> dr = cx.RadialVel(u.Quantity(1, "m/s"))
    >>> vec = dr * u.Quantity(2, "s")
    >>> vec
    RadialPos(r=Distance(value=f32[], unit=Unit("m")))
    >>> vec.r
    Distance(Array(2., dtype=float32), unit='m')

    >>> lax.mul(dr, u.Quantity(2, "s")).r
    Distance(Array(2., dtype=float32), unit='m')

    """
    return self.integral_cls.from_({k[2:]: v * other for k, v in field_items(self)})



"""Frames sub-package."""

__all__: list[str] = []

from typing import NoReturn

from plum import dispatch

from .base import AbstractReferenceFrame
from .errors import FrameTransformError
from .null import NoFrame


@dispatch(precedence=1)
def frame_transform_op(
    from_frame: NoFrame, to_frame: AbstractReferenceFrame, /
) -> NoReturn:
    """Cannot transform from the null frame.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> null = cxf.NoFrame()
    >>> icrs = cxf.ICRS()

    >>> try:
    ...     cxf.frame_transform_op(null, icrs)
    ... except cxf.FrameTransformError as e:
    ...     print(e)
    Cannot transform from the null frame.

    """
    msg = "Cannot transform from the null frame."
    raise FrameTransformError(msg)


@dispatch
def frame_transform_op(
    from_frame: AbstractReferenceFrame, to_frame: NoFrame, /
) -> NoReturn:
    """Cannot transform to the null frame.

    Examples
    --------
    >>> import coordinax.frames as cxf

    >>> null = cxf.NoFrame()
    >>> icrs = cxf.ICRS()

    >>> try:
    ...     cxf.frame_transform_op(icrs, null)
    ... except cxf.FrameTransformError as e:
    ...     print(e)
    Cannot transform to the null frame.

    """
    msg = "Cannot transform to the null frame."
    raise FrameTransformError(msg)



"""N-dimensional."""

__all__ = [
    # Base
    "AbstractPosND",
    "AbstractVelND",
    "AbstractAccND",
    # Cartesian
    "CartesianPosND",
    "CartesianVelND",
    "CartesianAccND",
    # Poincare
    "PoincarePolarVector",
]

# isort: split
from . import transform  # noqa: F401
from .base import AbstractAccND, AbstractPosND, AbstractVelND
from .cartesian import CartesianAccND, CartesianPosND, CartesianVelND
from .poincare import PoincarePolarVector



"""`coordinax.angle` private module.

Note that this module is private. Users should use the public API.

"""

__all__ = ["AbstractAngle", "Angle", "BatchableAngle", "BatchableAngleQ"]

from .base import AbstractAngle
from .core import Angle
from .typing import BatchableAngle, BatchableAngleQ

# isort: split
# Register the dispatching
from . import compat, register_primitives  # noqa: F401



"""Astronomy reference frames."""

__all__ = ["Galactocentric"]


from typing import ClassVar, TypeAlias, final

import equinox as eqx
from jaxtyping import Array, Shaped

import unxt as u

from coordinax._src.angle import Angle
from coordinax._src.distance import Distance
from coordinax._src.frames import AbstractReferenceFrame
from coordinax._src.vectors.d3 import CartesianVel3D, LonLatSphericalPos

ScalarAngle: TypeAlias = Shaped[u.Quantity["angle"] | Angle, ""]
RotationMatrix: TypeAlias = Shaped[Array, "3 3"]
LengthVector: TypeAlias = Shaped[u.Quantity["length"], "3"] | Shaped[Distance, "3"]
VelocityVector: TypeAlias = Shaped[u.Quantity["speed"], "3"]


@final
class Galactocentric(AbstractReferenceFrame):
    """Reference frame centered at the Galactic center.

    Based on the Astropy implementation of the Galactocentric frame.

    Examples
    --------
    >>> import coordinax as cx
    >>> frame = cx.frames.Galactocentric()
    >>> frame
    Galactocentric(
      galcen=LonLatSphericalPos( ... ),
      roll=Quantity[...](value=weak_i32[], unit=Unit("deg")),
      z_sun=Quantity[...](value=weak_f32[], unit=Unit("pc")),
      galcen_v_sun=CartesianVel3D( ... )
    )

    """

    #: RA, Dec, and distance of the Galactic center from an ICRS origin.
    #: ra, dec: https://ui.adsabs.harvard.edu/abs/2004ApJ...616..872R
    #: distance: https://ui.adsabs.harvard.edu/abs/2018A%26A...615L..15G
    galcen: LonLatSphericalPos = eqx.field(
        converter=LonLatSphericalPos.from_,
        default_factory=lambda: LonLatSphericalPos(
            lon=Angle(266.4051, "deg"),
            lat=Angle(-28.936175, "degree"),
            distance=Distance(8.122, "kpc"),
        ),
    )

    #: Rotation angle of the Galactic center from the ICRS x-axis.
    roll: u.Quantity["angle"] = eqx.field(
        converter=u.Quantity["angle"].from_, default=u.Quantity(0, "deg")
    )

    #: Distance from the Sun to the Galactic center.
    #: https://ui.adsabs.harvard.edu/abs/2019MNRAS.482.1417B
    z_sun: u.Quantity["length"] = eqx.field(
        converter=u.Quantity["length"].from_, default=u.Quantity(20.8, "pc")
    )

    #: Velocity of the Sun in the Galactic center frame.
    #: https://ui.adsabs.harvard.edu/abs/2018RNAAS...2..210D
    #: https://ui.adsabs.harvard.edu/abs/2018A%26A...615L..15G
    #: https://ui.adsabs.harvard.edu/abs/2004ApJ...616..872R
    galcen_v_sun: CartesianVel3D = eqx.field(
        converter=CartesianVel3D.from_,
        default_factory=lambda: CartesianVel3D.from_([12.9, 245.6, 7.78], "km/s"),
    )

    # --------

    #: The angle between the Galactic center and the ICRS x-axis.
    roll0: ClassVar[ScalarAngle] = u.Quantity(58.5986320306, "degree")



"""Built-in vector classes."""

__all__ = ["TwoSphereAcc", "TwoSpherePos", "TwoSphereVel"]

from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx

import unxt as u
from dataclassish.converters import Unless

import coordinax._src.typing as ct
from .base import AbstractAcc2D, AbstractPos2D, AbstractVel2D
from coordinax._src.angle import Angle, BatchableAngle
from coordinax._src.utils import classproperty
from coordinax._src.vectors.checks import check_polar_range
from coordinax._src.vectors.converters import converter_azimuth_to_range


# TODO: should this be named TwoSphericalPos or S(phere)(ical)2Pos
@final
class TwoSpherePos(AbstractPos2D):
    r"""Pos on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    theta : `coordinax.angle.Angle`
        Polar angle [0, 180] [deg] where 0 is the z-axis.
    phi : `coordinax.angle.Angle`
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.

    See Also
    --------
    `coordinax.SphericalPos`
        The counterpart in $R^3$, adding the polar distance coordinate $r$.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can construct a 2-spherical coordinate:

    >>> s2 = cx.TwoSpherePos(theta=u.Quantity(0, "deg"), phi=u.Quantity(180, "deg"))

    This coordinate has corresponding velocity class:

    >>> s2.differential_cls
    <class 'coordinax...TwoSphereVel'>

    """

    theta: BatchableAngle = eqx.field(converter=partial(Angle.from_, dtype=float))
    r"""Inclination angle :math:`\theta \in [0,180]`."""

    phi: BatchableAngle = eqx.field(
        converter=Unless(
            Angle, lambda x: converter_azimuth_to_range(Angle.from_(x, dtype=float))
        )
    )
    r"""Azimuthal angle :math:`\phi \in [0,360)`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        check_polar_range(self.theta)

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["TwoSphereVel"]:
        return TwoSphereVel


#####################################################################


@final
class TwoSphereVel(AbstractVel2D):
    r"""Vel on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    d_theta : Quantity['angular speed']
        Inclination speed $`d\theta/dt \in [-\infty, \infty]$.
    d_phi : Quantity['angular speed']
        Azimuthal speed $d\phi/dt \in [-\infty, \infty]$.

    See Also
    --------
    `coordinax.SphericalVel`
        The counterpart in $R^3$, adding the polar distance coordinate $d_r$.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can construct a 2-spherical velocity:

    >>> s2 = cx.TwoSphereVel(d_theta=u.Quantity(0, "deg/s"),
    ...                      d_phi=u.Quantity(2, "deg/s"))

    This coordinate has corresponding position and acceleration class:

    >>> s2.integral_cls
    <class 'coordinax...TwoSpherePos'>

    >>> s2.differential_cls
    <class 'coordinax...TwoSphereAcc'>

    """

    d_theta: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(u.Quantity["angular speed"].from_, dtype=float)
    )
    r"""Inclination speed :math:`d\theta/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(u.Quantity["angular speed"].from_, dtype=float)
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[TwoSpherePos]:
        return TwoSpherePos

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["TwoSphereAcc"]:
        return TwoSphereAcc


@final
class TwoSphereAcc(AbstractAcc2D):
    r"""Vel on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    d2_theta : Quantity['angular acceleration']
        Inclination acceleration $`d^2\theta/dt^2 \in [-\infty, \infty]$.
    d2_phi : Quantity['angular acceleration']
        Azimuthal acceleration $d^2\phi/dt^2 \in [-\infty, \infty]$.

    See Also
    --------
    `coordinax.SphericalAcc`
        The counterpart in $R^3$, adding the polar distance coordinate $d_r$.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    We can construct a 2-spherical acceleration:

    >>> s2 = cx.TwoSphereAcc(d2_theta=u.Quantity(0, "deg/s2"),
    ...                      d2_phi=u.Quantity(2, "deg/s2"))

    This coordinate has corresponding velocity class:

    >>> s2.integral_cls
    <class 'coordinax...TwoSphereVel'>

    """

    d2_theta: ct.BatchableAngularAcc = eqx.field(
        converter=partial(u.Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Inclination acceleration :math:`d^2\theta/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=partial(u.Quantity["angular acceleration"].from_, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[TwoSphereVel]:
        return TwoSphereVel



"""Radial 1D vectors."""

__all__ = ["RadialAcc", "RadialPos", "RadialVel"]

from functools import partial
from typing import final

import equinox as eqx
import jax
from plum import convert

import unxt as u
from dataclassish.converters import Unless

import coordinax._src.typing as ct
from .base import AbstractAcc1D, AbstractPos1D, AbstractVel1D
from coordinax._src.distance import AbstractDistance, BatchableDistance, Distance
from coordinax._src.utils import classproperty
from coordinax._src.vectors.checks import check_r_non_negative


@final
class RadialPos(AbstractPos1D):
    """Radial vector representation."""

    r: BatchableDistance = eqx.field(
        converter=Unless(AbstractDistance, partial(Distance.from_, dtype=float))
    )
    r"""Radial distance :math:`r \in [0,+\infty)`."""

    def __check_init__(self) -> None:
        """Check the initialization."""
        check_r_non_negative(self.r)

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["RadialVel"]:
        return RadialVel


@final
class RadialVel(AbstractVel1D):
    """Radial differential representation."""

    d_r: ct.BatchableSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""Radial speed :math:`dr/dt \in (-\infty,+\infty)`."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[RadialPos]:
        return RadialPos

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["RadialAcc"]:
        return RadialAcc

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        # TODO: change to UncheckedQuantity
        return jax.core.get_aval(convert(self, u.Quantity).value)


@final
class RadialAcc(AbstractAcc1D):
    """Radial differential representation."""

    d2_r: ct.BatchableAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""Radial acceleration :math:`d^2r/dt^2 \in (-\infty,+\infty)`."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[RadialVel]:
        return RadialVel

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        # TODO: change to UncheckedQuantity
        return jax.core.get_aval(convert(self, u.Quantity).value)



"""Astronomy reference frames."""

__all__ = ["ICRS"]


from typing import TypeAlias, final

from jaxtyping import Array, Shaped

import unxt as u

from coordinax._src.distance import Distance
from coordinax._src.frames import AbstractReferenceFrame

RotationMatrix: TypeAlias = Shaped[Array, "3 3"]
LengthVector: TypeAlias = Shaped[u.Quantity["length"], "3"] | Shaped[Distance, "3"]


@final
class ICRS(AbstractReferenceFrame):
    """The International Celestial Reference System (ICRS).

    Examples
    --------
    >>> import coordinax as cx
    >>> frame = cx.frames.ICRS()
    >>> frame
    ICRS()

    """



"""Base classes for distance quantities."""

__all__: list[str] = []

from abc import abstractmethod

from unxt.quantity import AbstractQuantity


class AbstractDistance(AbstractQuantity):  # type: ignore[misc]
    """Distance quantities."""

    @property
    @abstractmethod
    def distance(self) -> "Distance":
        """The distance."""

    @property
    @abstractmethod
    def parallax(self) -> "Parallax":
        """The parallax."""

    @property
    @abstractmethod
    def distance_modulus(self) -> "DistanceModulus":
        """The distance modulus."""



"""Mixin classes."""

__all__: list[str] = []

from collections.abc import Sequence
from typing import cast

import jax
from plum import convert

import unxt as u
from dataclassish import field_keys, field_values

from coordinax._src.vectors.funcs import represent_as


class AvalMixin:
    """Mixin class to add an ``aval`` method.

    See [quax doc](https://docs.kidger.site/quax/examples/custom_rules/) for
    more details.
    """

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        1 dimensional vectors:

        >>> vec = cx.CartesianPos1D.from_([1], "m")
        >>> vec.aval()
        ConcreteArray([1.], dtype=float32)

        >>> vec = cx.RadialPos.from_([1], "m")
        >>> vec.aval()
        ConcreteArray([1.], dtype=float32)

        >>> vec = cx.CartesianVel1D.from_([1], "m/s")
        >>> vec.aval()
        ConcreteArray([1], dtype=int32)

        >>> vec = cx.RadialVel.from_([1], "m/s")
        >>> vec.aval()
        ConcreteArray([1], dtype=int32)

        >>> vec = cx.CartesianAcc1D.from_([1], "m/s2")
        >>> vec.aval()
        ConcreteArray([1], dtype=int32)

        >>> vec = cx.RadialAcc.from_([1], "m/s2")
        >>> vec.aval()
        ConcreteArray([1], dtype=int32)

        2 dimensional vectors:

        >>> vec = cx.CartesianPos2D.from_([1, 2], "m")
        >>> vec.aval()
        ConcreteArray([1. 2.], dtype=float32)

        >>> vec = cx.PolarPos(r=u.Quantity(1, "m"), phi=u.Quantity(0, "rad"))
        >>> vec.aval()
        ConcreteArray([1. 0.], dtype=float32)

        >>> vec = cx.CartesianVel2D.from_([1, 2], "m/s")
        >>> vec.aval()
        ConcreteArray([1. 2.], dtype=float32)

        >>> vec = cx.PolarVel(d_r=u.Quantity(1, "m/s"), d_phi=u.Quantity(0, "rad/s"))
        >>> try: vec.aval()
        ... except NotImplementedError as e: print("nope")
        nope

        >>> vec = cx.CartesianAcc2D.from_([1,2], "m/s2")
        >>> vec.aval()
        ConcreteArray([1. 2.], dtype=float32)

        >>> vec = cx.PolarAcc(d2_r=u.Quantity(1, "m/s2"), d2_phi=u.Quantity(0, "rad/s2"))
        >>> try: vec.aval()
        ... except NotImplementedError as e: print("nope")
        nope

        3 dimensional vectors:

        >>> vec = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> vec.aval()
        ConcreteArray([1. 2. 3.], dtype=float32)

        >>> vec = cx.CartesianPos3D.from_([[1, 2, 3], [4, 5, 6]], "m")
        >>> vec.aval()
        ConcreteArray([[1. 2. 3.]
                       [4. 5. 6.]], dtype=float32)

        >>> vec = cx.SphericalPos(r=u.Quantity(1, "m"), phi=u.Quantity(0, "rad"), theta=u.Quantity(0, "rad"))
        >>> vec.aval()
        ConcreteArray([0. 0. 1.], dtype=float32)

        >>> vec = cx.CartesianVel3D.from_([1,2,3], "m/s")
        >>> vec.aval()
        ConcreteArray([1. 2. 3.], dtype=float32)

        >>> vec = cx.SphericalVel(d_r=u.Quantity(1, "m/s"), d_phi=u.Quantity(0, "rad/s"), d_theta=u.Quantity(0, "rad/s"))
        >>> try: vec.aval()
        ... except NotImplementedError as e: print("nope")
        nope

        >>> vec = cx.CartesianAcc3D.from_([1,2,3], "m/s2")
        >>> vec.aval()
        ConcreteArray([1. 2. 3.], dtype=float32)

        >>> vec = cx.SphericalAcc(d2_r=u.Quantity(1, "m/s2"), d2_phi=u.Quantity(0, "rad/s2"), d2_theta=u.Quantity(0, "rad/s2"))
        >>> try: vec.aval()
        ... except NotImplementedError as e: print("nope")
        nope

        """  # noqa: E501
        # TODO: change to UncheckedQuantity
        target = self._cartesian_cls  # type: ignore[attr-defined]
        return jax.core.get_aval(convert(represent_as(self, target), u.Quantity).value)


##############################################################################

SUPPORTED_IPYTHON_REPR_FORMATS: dict[str, str] = {
    "text/plain": "__repr__",
    "text/latex": "_repr_latex_",
}


class IPythonReprMixin:
    """Mixin class to add IPython rich display methods."""

    def _repr_mimebundle_(
        self,
        *,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
    ) -> dict[str, str]:
        r"""Return a MIME bundle representation of the Quantity.

        Parameters
        ----------
        include, exclude : Sequence[str] | None, optional
            The set of keys to include / exclude in the MIME bundle. If not
            provided, all supported formats are included. 'include' has
            precedence over 'exclude'.

        Examples
        --------
        >>> import coordinax as cx

        >>> vec = cx.CartesianPos2D.from_([1, 2], "m")
        >>> vec._repr_mimebundle_()
        {'text/plain': 'CartesianPos2D(\n  x=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m")),\n  y=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m"))\n)',
         'text/latex': '$\\left( \\begin{matrix}\\mathrm{ x } \\\\ \\mathrm{ y }\\end{matrix} \\right)=\\left( \\begin{matrix}1. \\; \\mathrm{m} \\\\ 2. \\; \\mathrm{m}\\end{matrix} \\right)$'}

        >>> vec._repr_mimebundle_(include=["text/plain"])
        {'text/plain': 'CartesianPos2D(\n  x=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m")),\n  y=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m"))\n)'}

        >>> vec._repr_mimebundle_(exclude=["text/latex"])
        {'text/plain': 'CartesianPos2D(\n  x=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m")),\n  y=Quantity[PhysicalType(\'length\')](value=f32[], unit=Unit("m"))\n)'}

        """  # noqa: E501
        # Determine the set of keys to include in the MIME bundle
        keys: Sequence[str]
        if include is None and exclude is None:
            keys = tuple(SUPPORTED_IPYTHON_REPR_FORMATS)
        elif include is not None:
            keys = [key for key in include if key in SUPPORTED_IPYTHON_REPR_FORMATS]
        else:
            keys = [
                k for k in SUPPORTED_IPYTHON_REPR_FORMATS if k not in cast(str, exclude)
            ]

        # Create the MIME bundle
        return {
            key: getattr(self, SUPPORTED_IPYTHON_REPR_FORMATS[key])() for key in keys
        }

    def _repr_latex_(self) -> str:
        r"""Return a LaTeX representation of the Quantity.

        Examples
        --------
        >>> import coordinax as cx

        >>> vec = cx.CartesianPos2D.from_([1, 2], "m")
        >>> vec._repr_latex_()
        '$\\left( \\begin{matrix}\\mathrm{ x } \\\\ \\mathrm{ y }\\end{matrix} \\right)=\\left( \\begin{matrix}1. \\; \\mathrm{m} \\\\ 2. \\; \\mathrm{m}\\end{matrix} \\right)$'

        """  # noqa: E501
        # TODO: better latex representation of the components. Currently
        # velocities are shown as `dx` and accelerations as `d2x`.
        ks = (r"\mathrm{ " + k.replace("_", "") + r" }" for k in field_keys(self))
        latex_ks = r"\left( \begin{matrix}" + r" \\ ".join(ks) + r"\end{matrix} \right)"

        vs = (v._repr_latex_()[1:-1] for v in field_values(self))
        latex_vs = r"\left( \begin{matrix}" + r" \\ ".join(vs) + r"\end{matrix} \right)"

        return r"$" + latex_ks + "=" + latex_vs + r"$"



"""Type hints for `coordinax.angle`."""

__all__: list[str] = []


from jaxtyping import Shaped

import unxt as u

from .base import AbstractAngle

#: Batchable angular-type Quantity.
BatchableAngularQuantity = Shaped[u.Quantity["angle"], "*#batch"]

#: Batchable Angle.
BatchableAngle = Shaped[AbstractAngle, "*#batch"]

#: Batchable Angle or Angular Quantity.
BatchableAngleQ = BatchableAngle | BatchableAngularQuantity
